import typing

from iterpop import iterpop as ip
import opytional as opyt
from safe_assert import safe_assert as always_assert

from ...._detail import PolicyCouplerBase
from ..._PolicySpec import PolicySpec
from ..._impl import calc_rank_sep, get_retained_ranks


class GenDropRanks:
    """Functor to implement the approximate space-filling MRCA-recency-
    proportional resolution stratum retention policy, for use with
    HereditaryStratigraphicColumn.

    This functor enacts the approximate space-filling MRCA-recency-
    proportional resolution stratum retention policy by specifying the set of
    strata ranks that should be purged from a hereditary stratigraphic column
    when the nth stratum is deposited.
    """

    # store the most recent result from calls to __call__ that dropped a rank
    # in the format
    #
    # (num_stratum_depositions_completed, dropped_rank)
    #
    # this cache enables optimization if calls to __call__ are made in
    # monotonically increasing order with respect to
    # num_stratum_depositions_completed but is not necessary for __call__ to
    # produce a correct result
    # (if the optimization is not available, dropped rank is computed using a
    # slower fallback method)
    #
    # note that because this policy maintains column size exactly at a fixed
    # size bound so at most only one rank will be dropped when a new stratum is
    # deposited
    # (zero ranks are dropped while the column is growing to that size bound)
    _cached_result: typing.Optional[typing.Tuple[int, int]]

    def __init__(
        self: "GenDropRanks",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        self._cached_result = None

    def __eq__(self: "GenDropRanks", other: typing.Any) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "GenDropRanks",
        policy: PolicyCouplerBase,
        num_stratum_depositions_completed: int,
        retained_ranks: typing.Optional[typing.Iterable[int]],
    ) -> typing.Iterator[int]:
        """Decide which strata within the stratagraphic column should be purged.

        Every time a new stratum is deposited, this method is called to
        determine which strata should be purged. All strata at ranks yielded
        from this functor are immediately purged from the column, meaning that
        for a stratum to persist it must not be yielded by this functor each
        and every time a new stratum is deposited.

        Parameters
        ----------
        policy: Policy
            Policy this functor enacts.
        num_stratum_depositions_completed : int
            The number of strata that have already been deposited, not
            including the latest stratum being deposited which prompted the
            current purge operation.
        retained_ranks : iterator over int
            An iterator over ranks of strata currently retained within the
            hereditary stratigraphic column. Not used in this functor.

        Returns
        -------
        iterator over int
                The ranks of strata that should be purged from the hereditary
            stratigraphic column at this deposition step.

        See Also
        --------
        depth_proportional_resolution_tapered_algo:
            For details on the rationale, implementation, and guarantees of the
            tapered depth-proportional resolution stratum retention policy.
        """
        spec = policy.GetSpec()

        size_bound = policy.CalcNumStrataRetainedUpperBound(
            num_stratum_depositions_completed,
        )

        # try a series of shortcuts that only apply in certain cases
        # if no shortcut cases are available (and res is still None), fallback
        # to calculating res as the difference between cur retained ranks and
        # prev retained ranks
        res = None

        # if we haven't filled up available space, no ranks are dropped
        if num_stratum_depositions_completed < size_bound:
            res = set()
        # if we have a cached result for the current time point, use it
        elif opyt.apply_if(
            self._cached_result,
            lambda x: x[0] == num_stratum_depositions_completed,
        ):
            cached_time, cached_drop = self._cached_result
            res = {
                cached_drop,
            }
        # if degree > 0 and we have a cached result for the preceding time point
        # try to use shortcut to return same result for current time point
        elif spec.GetDegree() and opyt.apply_if(
            self._cached_result,
            lambda x: x[0] == num_stratum_depositions_completed - 1,
        ):
            # shortcut explanation...
            #
            # recall that the geom seq nth root policy is defined by a fixed
            # degree d and providing guaranteed coverage to d + 1 targets
            #
            # let k range from 0 to d, inclusive and n represent the number of
            # stratum depositions elapsed... then each target tracks rank
            #
            # n - (n)^(k/d)
            #
            # the tapered geom seq nth root policy extends the above by
            # retaining additional ranks to ensure that a fixed size is
            # maintained
            #
            # the tapered geom seq nth root policy accomplishes this by
            # creating an iterator over priority ranks for each kth target and
            # drawing round robin from these iterators
            #
            # each iterator is drawn from until a novel (not yet decided for
            # retention) rank is yielded, then the next iterator is drawn from
            #
            # if an iterator is exhausted, it is skipped over (so the set of
            # remaining iterators will be polled an additional rank than they
            # would not have had to produce otherwise)
            #
            # note that the pool of retained ranks being drawn into is seeded
            # with rank 0 and rank n - 1, because these must be retained
            #
            # note also that targets' iterators are drawn from in the order of
            # most ancient to most recent (i.e., k = d down to k = 1)
            #
            # in order to make this shortcut optimization more effective and
            # simplify its implementation, the k = 0 iterator is excluded from
            # this phase
            #
            # only after (and if) all k > 0 iterators have been exhausted is
            # the k = 0 iterator drawn from
            #
            # the k = 0 iterator is the lowest-priority iterator that
            # exclusively vacuums up (i.e., yields) ranks that have been
            # yielded from no other iterator... put another way, the k = 0
            # iterator will never pre-empt another iterator by claiming a rank
            # "first"
            #
            #
            # this optimization focuses on the k = 0 iterator, which
            # corresponds to the target that tracks the most recent rank
            # deposited
            #
            # because the k = 0 target tracks the most recent rank, the k = 0
            # iterator yields all deposited ranks in reverse order
            # i.e., this iterator yields reversed(range(n))
            #
            # as n increases, more and more of dropped ranks will correspond
            # to the most ancient rank yielded from the 0th iterator
            # (considering the implicit seed n - 1 as yielded from this
            # iterator)
            #
            # under an ideal scenario where no iterators are exhausted the 0th
            # iterator will yield just the second-to-last rank (i.e., n - 2)
            # because the last rank is always retained
            #
            # however, if other iterators are exhausted, the 0th iterator may
            # be polled for more ranks
            #
            # only considering cases where no other iterators are exhausted
            # a) is hard enough computationally to probably negate opt benefits
            # b) reduces the cases where this optimization applies (i.e.,
            #    excludes optimization at such timepoints)
            #
            # that raises the question, how to compute the number of items
            # yielded from the k = 0 iterator?
            #
            # we will employ a small amount of caching to sidestep this
            # problem: we can look at the number of items yielded from the
            # k = 0 iterator at the previous timepoint
            #
            # our shortcut optimization will rely on detecting whether
            # 1) there could possibly have been any change to the number of
            #    items yielded from the k = 0 iterator due to a new highest-
            #    priority rank becoming available to any other k > 1 iterator
            #    (which would cause the to dropped rank jump forward more than
            #    one position) (every other iterator gets first dibs above the
            #    k = 0 iterator, so this is the only situation may result in a
            #    k > 0 iterator's lowest-priority rank getting dropped)
            # 2) there could possibly be a latent rank being retained by a
            #    k > 0 iterator at the end of the chain of k = 0's retain ranks
            #    that the k = 0 iterator's retained rainks had previously been
            #    stepping over (thereby extending the k = 0 iterator's most
            #    ancient retained rank backward) (and might cause the dropped
            #    rank to jump forward more than one position)
            #
            # if neither of these scenarios is possible, we can guarantee that
            # the next dropped rank will immediately succeed the previous
            # (cached) dropped rank
            #
            # scenario 1 is only possible when the deposited rank could align
            # with the current sep (rank-to-rank step size) of the k = 1
            # iterator
            # (all iterators k > 1 have monotonically increasing seps that only
            # increase by doubling so are covered by checking the k = 1 case)
            #
            # scenario 2 is only possible when the rank-to-be-dropped could
            # have been retained as a priority rank of a k > 1 iterator...
            # so, check whether the contemporaneous to rank-to-be-dropped k = 1
            # sep could have aligned with that rank
            # (again, all iterators k > 1 have monotonically increasing seps
            # that only increase by doubling so are covered by checking the
            # k = 1 case)
            #
            # NOTE: on further reflection, it seems that scenario 2 is the only
            # relevant scenario because the newly-added rank is always
            # guaranteed to be retained due to the pool of retained ranks being
            # seeded with it... have disabled unnecessary restriction below

            # unpack cache
            cached_time, cached_drop_rank = self._cached_result
            assert cached_drop_rank > 0

            # what is the tightest spacing between retained ranks for
            # any non-0th pow iterator at the current timepoint?
            # allows us to detect whether deposited rank is relevant to any
            # k > 1 iterator
            pow1_newsep = calc_rank_sep(
                spec.GetDegree(),
                spec.GetInterspersal(),
                1,
                # no +1 to be conservative
                num_stratum_depositions_completed,
            )
            # what is the tightest spacing between retained ranks for
            # any non-0th pow iterator at the end of the k = 0 retention chain?
            # (i.e., at rank-to-be-dropped)
            # allows us to detect any possible ranks retained for k > 0
            # that may be retained even if the apparent k = 0 retention chain
            # advances past them
            # (this would interfere with our assumption of the k = 0 drop
            # moving forward one rank per deposition)
            pow1_oldsep = calc_rank_sep(
                spec.GetDegree(),
                spec.GetInterspersal(),
                1,
                # no +1 to be conservative
                cached_drop_rank,
            )
            assert pow1_oldsep <= pow1_newsep

            if (
                # scenario 1 (NOT ACTUALLY NECESSARY, DISABLED)
                # ensure deposited rank is irrelevant to all pow > 0
                # no +1 due to translation from len to index of last
                #
                # num_stratum_depositions_completed % pow1_newsep > 1 and
                #
                # scenario 2
                # ensure rank-to-drop isn't possibly retained for pow > 0
                (cached_drop_rank + 1) % pow1_oldsep
                > 1
                # ^^^ note that >1's ensure that we are at least two ranks
                # past any problem rank (i.e., that the preceding call
                # computed a drop_rank at the end of the k = 0 retention chain)
            ):
                # sufficient guarantee drop moves forward exactly 1
                res = {
                    # drop rank 1 past the one we last dropped
                    cached_drop_rank
                    + 1,
                }

        elif spec.GetDegree():
            # TODO case with no caching?
            # can we prove when pow0 retention chain is only 1 long?
            # i.e., previous iterators don't exhaust?
            pass

        prev_retained_ranks = get_retained_ranks(
            policy,
            num_stratum_depositions_completed,
        )

        cur_retained_ranks = get_retained_ranks(
            policy,
            num_stratum_depositions_completed + 1,
        )

        if res is not None:
            always_assert(
                res == (prev_retained_ranks - cur_retained_ranks),
                {
                    "res": res,
                    "(prev_retained_ranks - cur_retained_ranks)": (
                        prev_retained_ranks - cur_retained_ranks
                    ),
                },
            )
        else:
            res = prev_retained_ranks - cur_retained_ranks

        # # have we determined res using a shortcut method?
        # # if not, fall back to do full computation for res
        # if res is None:
        #     prev_retained_ranks = super(
        #         StratumRetentionCondemnerTaperedGeomSeqNthRoot,
        #         self,
        #     )._get_retained_ranks(
        #         num_stratum_depositions_completed,
        #     )
        #
        #     cur_retained_ranks = super(
        #         StratumRetentionCondemnerTaperedGeomSeqNthRoot,
        #         self,
        #     )._get_retained_ranks(
        #         num_stratum_depositions_completed + 1,
        #     )
        #     # take set difference between prev retained ranks and cur retained
        #     # ranks
        #     res = (prev_retained_ranks - cur_retained_ranks)

        # if res is not empty, it will only have one rank
        # cache that dropped rank with the current timepoint
        if res:
            self._cached_result = (
                num_stratum_depositions_completed,
                ip.popsingleton(res),
            )

        yield from res



import numpy as np
import packaging


def numpy_fromiter_polyfill(iter_: iter, dtype) -> np.ndarray:
    """An implementation of `numpy.fromiter` that polyfills support for object
    dtype introduced in numpy version 1.23.

    If numpy is older than version 1.23, `iter_` will be unpacked into a list
    and the `np.array` constructor will be used. Otherwise, `numpy.fromiter`
    will be used.

    Optional arguments `count` and `like` are not supported.

    Parameters
    ----------
    iter_ : iterable
        An iterable object providing data for the array.
    dtype :
        The data-type of the returned array.

    Returns
    -------
    numpy.ndarray
        A numpy array containing the elements of the iterable.

    Examples
    --------
    >>> a = [1, 2, 3, 4, 5]
    >>> b = numpy_fromiter_polyfill(a, np.int32)
    >>> b
    array([1, 2, 3, 4, 5])
    """
    if (
        packaging.version.parse(np.__version__)
        < packaging.version.parse("1.23")
        and np.object_ == dtype
    ):
        return np.array([*iter_], dtype=dtype)
    else:
        return np.fromiter(
            iter_,
            dtype=dtype,
        )



import os
import typing

import pandas as pd
import polars as pl
import pytest

from hstrat._auxiliary_lib import (
    coerce_to_polars,
    delegate_polars_implementation,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@delegate_polars_implementation(recurse_type_checks=True)
def dummy_func(
    df_dict: dict[str, pd.DataFrame],
    series_list_singleton: typing.List[pd.Series],
    dummy_val: typing.Any,
) -> typing.Tuple[pd.DataFrame, pd.Series, typing.Any]:
    assert isinstance(df_dict["df"], pd.DataFrame) and isinstance(
        series_list_singleton[0], pd.Series
    )
    return df_dict["df"], series_list_singleton[0], dummy_val


@pytest.mark.parametrize(
    "df",
    [
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
        coerce_to_polars(pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv")),
        coerce_to_polars(
            pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv")
        ),
        coerce_to_polars(
            pd.read_csv(f"{assets_path}/nk_tournamentselection.csv")
        ),
    ],
)
@pytest.mark.parametrize(
    "series",
    [
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv")["id"],
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv")["origin_time"],
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv")["phenotype"],
        coerce_to_polars(
            pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv")["id"]
        ),
        coerce_to_polars(
            pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv")[
                "origin_time"
            ]
        ),
        coerce_to_polars(
            pd.read_csv(f"{assets_path}/nk_tournamentselection.csv")[
                "phenotype"
            ]
        ),
    ],
)
def test_coercion_and_error(
    df: typing.Union[pd.DataFrame, pl.DataFrame],
    series: typing.Union[pd.Series, pl.Series],
):
    if isinstance(df, pd.DataFrame) ^ isinstance(series, pd.Series):
        with pytest.raises(
            TypeError, match="mixing pandas and polars types is disallowed"
        ):
            dummy_func({"df": df}, [series], 1234)
    else:
        new_df, new_series, _ = dummy_func({"df": df}, [series], "asdf")
        assert type(new_df) == type(df)
        assert type(new_series) == type(series)


SignalException = type("", (Exception,), {})


def polars_func(df: pl.DataFrame):
    assert isinstance(df, pl.DataFrame)
    raise SignalException("Using polars function")


@delegate_polars_implementation(polars_func)
def pandas_func(df: pd.DataFrame):
    assert isinstance(df, pd.DataFrame)
    return df


@pytest.mark.parametrize(
    "df",
    [
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
        coerce_to_polars(pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv")),
        coerce_to_polars(
            pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv")
        ),
        coerce_to_polars(
            pd.read_csv(f"{assets_path}/nk_tournamentselection.csv")
        ),
    ],
)
def test_delegation(df: typing.Union[pd.DataFrame, pl.DataFrame]) -> None:
    if isinstance(df, pl.DataFrame):
        with pytest.raises(SignalException, match="Using polars function"):
            pandas_func(df)
    else:
        assert isinstance(pandas_func(df), pd.DataFrame)



import itertools as it
import typing

from iterpop import iterpop as ip
import numpy as np
import pandas as pd

from .._auxiliary_lib import as_nullable_type
from ._HereditaryStratigraphicAssemblageSpecimen import (
    HereditaryStratigraphicAssemblageSpecimen,
)
from ._HereditaryStratigraphicSpecimen import HereditaryStratigraphicSpecimen


class HereditaryStratigraphicAssemblage:
    """A collection of HereditaryStratigraphicSpecimens, padded to include entries for all ranks retained by any specimen within the
    assemblage.

    This allows for more efficient comparisons between specimens, due to direct
    alignment.

    Parameters
    ----------
    specimens : iterable of HereditaryStratigraphicSpecimen
        The specimens that make up the assemblage.

    See Also
    --------
    HereditaryStratigraphicSpecimen
        Type alias for a postprocessing representation of the differentia
        retained by an extant HereditaryStratigraphicColumn, indexed by
        deposition rank.
    assemblage_from_records
        Deserialize a `HereditaryStratigraphicSpecimen` from a dict composed of
        builtin data types.
    pop_to_assemblage
        Create a `HereditaryStratigraphicAssemblage` from a collection of
        `HereditaryStratigraphicColumn`s.
    """

    __slots__ = ("_assemblage_df", "_stratum_differentia_bit_width")

    # rows indexed by rank
    # columns represent individuals within population, as IntegerArrays
    _assemblage_df: pd.DataFrame

    _stratum_differentia_bit_width: int

    def __init__(
        self: "HereditaryStratigraphicAssemblage",
        specimens: typing.Iterable[HereditaryStratigraphicSpecimen],
    ) -> None:
        """Construct a new HereditaryStratigraphicAssemblage instance.

        Takes a collection of HereditaryStratigraphicSpecimen instances and
        creates a new HereditaryStratigraphicAssemblage instance containing
        those specimens.
        """
        specimens1, specimens2 = it.tee(specimens)
        self._stratum_differentia_bit_width = ip.pourhomogeneous(
            specimen.GetStratumDifferentiaBitWidth() for specimen in specimens1
        )
        try:
            self._assemblage_df = pd.concat(
                (
                    as_nullable_type(specimen.GetData())
                    for specimen in specimens2
                ),
                axis="columns",
            ).sort_index()
        except ValueError:  # empty specimens
            self._assemblage_df = pd.DataFrame()

        assert not self._assemblage_df.index.isna().any()
        self._assemblage_df.index.astype(np.uint64, copy=False)

    def BuildSpecimens(
        self: "HereditaryStratigraphicAssemblage",
    ) -> typing.Iterator[HereditaryStratigraphicAssemblageSpecimen]:
        """Iterator over specimens in assemblage as potentially-padded
        HereditaryStratigraphicAssemblageSpecimen objects."""
        for _column_name, series in self._assemblage_df.items():
            yield HereditaryStratigraphicAssemblageSpecimen(
                stratum_differentia_series=series,
                stratum_differentia_bit_width=(
                    self._stratum_differentia_bit_width
                ),
            )



import numpy as np

from hstrat._auxiliary_lib import numpy_index_flat


def test_numpy_index_flat_first_occurrence():
    # Test that the function returns the correct index for the first occurrence of an item
    a = np.array([1, 2, 3, 4, 5])
    assert numpy_index_flat(a, 3) == 2
    assert a[numpy_index_flat(a, 3)] == 3
    assert numpy_index_flat(a, 5) == 4
    assert a[numpy_index_flat(a, 5)] == 5
    assert numpy_index_flat(a, 1) == 0
    assert a[numpy_index_flat(a, 1)] == 1

    b = np.array([0, 1, 2, 1, 4])
    assert numpy_index_flat(b, 1) == 1
    assert b[numpy_index_flat(b, 1)] == 1

    c = np.array([1])
    assert numpy_index_flat(c, 1) == 0
    assert c[numpy_index_flat(c, 1)] == 1

    c = np.array([1])
    assert numpy_index_flat(c, 1, n=-1) == 0
    assert c[numpy_index_flat(c, 1, n=-1)] == 1


def test_numpy_index_flat_reverse_index():
    # Test that the function returns the correct index for a reverse indexed array
    a = np.array([5, 4, 3, 2, 1])
    assert numpy_index_flat(a[::-1], 3) == 2


def test_numpy_index_flat_item_not_found():
    # Test that the function returns None when the item is not found in the array
    a = np.array([1, 2, 3, 4, 5])
    assert numpy_index_flat(a, 6) is None

    b = np.array([0, 1, 2, 1, 4])
    assert numpy_index_flat(b, 3) is None

    c = np.array([1])
    assert numpy_index_flat(c, 2) is None


def test_numpy_index_flat_empty_array():
    # Test that the function returns None when the array is empty
    a = np.array([])
    assert numpy_index_flat(a, 1) is None
    assert numpy_index_flat(a, 1, n=-1) is None


def test_numpy_index_flat_reverse():
    # Test that the function returns the correct index for a reverse indexed array
    a = np.array([1, 2, 3, 4, 5])
    assert numpy_index_flat(a, 3, n=-1) == 2
    assert a[numpy_index_flat(a, 3, n=-1)] == 3
    assert numpy_index_flat(a, 6, n=-1) is None
    assert numpy_index_flat(a, 1, n=-1) == 0
    assert a[numpy_index_flat(a, 1, n=-1)] == 1

    b = np.array([0, 1, 2, 1, 4])
    assert numpy_index_flat(b, 1, n=-1) == 3
    assert b[numpy_index_flat(b, 1, n=-1)] == 1
    assert numpy_index_flat(b, 3, n=-1) is None

    c = np.array([1])
    assert numpy_index_flat(c, 1, n=-1) == 0
    assert c[numpy_index_flat(c, 1, n=-1)] == 1
    assert numpy_index_flat(c, 2, n=-1) is None


def test_numpy_index_flat_n0():
    a = np.array([1, 2, 3, 4, 5, 3, 3])
    b = np.array([1, 2, 4, 5])
    c = np.array([])
    assert numpy_index_flat(a, 3, n=0) == 2
    assert a[numpy_index_flat(a, 3, n=0)] == 3

    assert numpy_index_flat(b, 3, n=0) is None
    assert numpy_index_flat(c, 3, n=0) is None


def test_numpy_index_flat_n1():
    a = np.array([1, 2, 3, 4, 5, 3, 3])
    b = np.array([1, 2, 4, 5])
    c = np.array([])
    assert numpy_index_flat(a, 3, n=1) == 5
    assert a[numpy_index_flat(a, 3, n=1)] == 3

    assert numpy_index_flat(b, 3, n=1) is None
    assert numpy_index_flat(c, 3, n=1) is None


def test_numpy_index_flat_n2():
    a = np.array([1, 2, 3, 4, 5, 3, 3])
    b = np.array([1, 2, 4, 5])
    c = np.array([])
    assert numpy_index_flat(a, 3, n=2) == 6
    assert a[numpy_index_flat(a, 3, n=2)] == 3

    assert numpy_index_flat(b, 3, n=2) is None
    assert numpy_index_flat(c, 3, n=2) is None


def test_numpy_index_flat_nneg1():
    a = np.array([1, 2, 3, 4, 5, 3, 3])
    b = np.array([1, 2, 4, 5])
    c = np.array([])
    assert numpy_index_flat(a, 3, n=-1) == 6
    assert a[numpy_index_flat(a, 3, n=-1)] == 3

    assert numpy_index_flat(b, 3, n=-1) is None
    assert numpy_index_flat(c, 3, n=-1) is None


def test_numpy_index_flat_nneg2():
    a = np.array([1, 2, 3, 4, 5, 3, 3])
    b = np.array([1, 2, 4, 5])
    c = np.array([])
    assert numpy_index_flat(a, 3, n=-2) == 5
    assert a[numpy_index_flat(a, 3, n=-2)] == 3

    assert numpy_index_flat(b, 3, n=-2) is None
    assert numpy_index_flat(c, 3, n=-2) is None



import pandas as pd


def alifestd_is_chronologically_sorted(
    phylogeny_df: pd.DataFrame,
    how: str = "origin_time",
) -> bool:
    """Do rows appear in chronological order?

    Defaults to `origin_time`. Input dataframe is not mutated by this operation.
    """

    return how in phylogeny_df and phylogeny_df[how].is_monotonic_increasing



import collections
import itertools as it
import typing

from ._consume import consume


# adapted from https://stackoverflow.com/a/11443871/17332200
def splicewhile(
    predicate: typing.Callable, iterator: typing.Iterator
) -> typing.Tuple[typing.List, typing.Iterator]:
    """Create list of iterator items before predicate is false, consuming them
    from the iterator."""

    if not isinstance(iterator, collections.abc.Iterator):
        # consume is only effectual on an iterator, not an iterable
        # enforce against misconceived usage
        raise TypeError

    it1, it2 = it.tee(iterator)
    res = [*it.takewhile(predicate, it1)]
    consume(it2, len(res))
    return res, it2



import numpy as np
import pandas as pd

from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_is_topologically_sorted import alifestd_is_topologically_sorted
from ._alifestd_topological_sort import alifestd_topological_sort
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col
from ._jit import jit


@jit(nopython=True)
def _alifestd_calc_node_depth_asexual_contiguous(
    ancestor_ids: np.ndarray,
) -> np.ndarray:
    """Optimized implementation for asexual phylogenies with contiguous ids."""
    ancestor_ids = ancestor_ids.astype(np.uint64)
    node_depths = np.full_like(ancestor_ids, -1, dtype=np.int64)

    for id_, _ in enumerate(ancestor_ids):
        ancestor_id = ancestor_ids[id_]
        ancestor_depth = node_depths[ancestor_id]
        node_depths[id_] = ancestor_depth + 1

    return node_depths


def alifestd_mark_node_depth_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """Add column `node_depth`, counting the number of nodes between a node
    and the root.

    A topological sort will be applied if `phylogeny_df` is not topologically
    sorted. Reindexing may be applied.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    if not alifestd_is_topologically_sorted(phylogeny_df):
        phylogeny_df = alifestd_topological_sort(phylogeny_df, mutate=True)

    if alifestd_has_contiguous_ids(phylogeny_df):
        # optimized implementation for contiguous ids
        node_depths = _alifestd_calc_node_depth_asexual_contiguous(
            phylogeny_df["ancestor_id"].values,
        )
        phylogeny_df["node_depth"] = node_depths
        return phylogeny_df

    # slower fallback implementation
    phylogeny_df.index = phylogeny_df["id"]

    phylogeny_df["node_depth"] = -1

    for idx in phylogeny_df.index:
        ancestor_id = phylogeny_df.loc[idx, "ancestor_id"]
        ancestor_depth = phylogeny_df.loc[ancestor_id, "node_depth"]
        phylogeny_df.loc[idx, "node_depth"] = ancestor_depth + 1

    return phylogeny_df



import typing

from .._auxiliary_lib import HereditaryStratigraphicArtifact


def diff_retained_ranks(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
) -> typing.Tuple[typing.Set[int], typing.Set[int]]:
    """Return ranks retained by first but not second, and vice versa.

    Returned as a tuple of sets.
    """
    first_ranks = set(first.IterRetainedRanks())
    second_ranks = set(second.IterRetainedRanks())

    return (
        first_ranks - second_ranks,
        second_ranks - first_ranks,
    )



import typing

from ...._detail import PolicyCouplerBase
from ..._PolicySpec import PolicySpec
from ..._impl import calc_provided_uncertainty


class GenDropRanks:
    """Functor to implement the tapered depth-proportional resolution stratum
    retention policy, for use with HereditaryStratigraphicColumn.

    This functor enacts the tapered depth-proportional resolution policy by
    specifying the set of strata ranks that should be purged from a hereditary
    stratigraphic column when the nth stratum is deposited.
    """

    def __init__(
        self: "GenDropRanks",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(self: "GenDropRanks", other: typing.Any) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "GenDropRanks",
        policy: PolicyCouplerBase,
        num_stratum_depositions_completed: int,
        retained_ranks: typing.Optional[typing.Iterable[int]],
    ) -> typing.Iterator[int]:
        """Decide which strata within the stratagraphic column should be purged.

        Every time a new stratum is deposited, this method is called to
        determine which strata should be purged. All strata at ranks yielded
        from this functor are immediately purged from the column, meaning that
        for a stratum to persist it must not be yielded by this functor each
        and every time a new stratum is deposited.

        Parameters
        ----------
        policy: Policy
            Policy this functor enacts.
        num_stratum_depositions_completed : int
            The number of strata that have already been deposited, not
            including the latest stratum being deposited which prompted the
            current purge operation.
        retained_ranks : iterator over int
            An iterator over ranks of strata currently retained within the
            hereditary stratigraphic column. Not used in this functor.

        Returns
        -------
        iterator over int
                The ranks of strata that should be purged from the hereditary
            stratigraphic column at this deposition step.

        See Also
        --------
        depth_proportional_resolution_tapered_algo:
            For details on the rationale, implementation, and guarantees of the
            tapered depth-proportional resolution stratum retention policy.
        """
        spec = policy.GetSpec()
        resolution = spec.GetDepthProportionalResolution()

        # until sufficient strata have been deposited to reach target resolution
        # don't remove any strata
        if num_stratum_depositions_completed < 2 * resolution + 1:
            return

        cur_stage_uncertainty = calc_provided_uncertainty(
            resolution, num_stratum_depositions_completed
        )
        prev_stage_uncertainty = cur_stage_uncertainty // 2
        assert prev_stage_uncertainty

        prev_stage_idx = (
            num_stratum_depositions_completed - 1
        ) // prev_stage_uncertainty

        second_newest_stratum_rank = num_stratum_depositions_completed - 1
        # we just added a new peg so we have to clear out an old one
        if second_newest_stratum_rank % prev_stage_uncertainty == 0:

            target_idx = prev_stage_idx * 2 - 4 * resolution + 1
            target_rank = target_idx * prev_stage_uncertainty
            if 0 < target_rank < num_stratum_depositions_completed:
                yield target_rank

        else:
            # newest stratum is in-progress deposition
            # that will occupy rank num_stratum_depositions_completed
            # we always keep the newest stratum
            # but because we know the second newest is not needed as a waypoint
            # we will get rid of it
            yield second_newest_stratum_rank



import polars as pl

from ._alifestd_make_ancestor_list_col_polars import (
    alifestd_make_ancestor_list_col_polars,
)


def alifestd_try_add_ancestor_list_col_polars(
    phylogeny_df: pl.DataFrame,
    root_ancestor_token: str = "none",
    mutate: bool = False,
) -> pl.DataFrame:
    """Add an ancestor_list column to the input DataFrame if the column does
    not already exist.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.

    Notes
    -----
    Even allowed by `mutate` flag, no side effects occur on input dataframe
    under Polars implementation. Flag is included for API compatibility with
    Pandas implementation.

    See Also
    --------
    alifestd_make_ancestor_list_col
    """
    if "ancestor_id" in phylogeny_df and "ancestor_list" not in phylogeny_df:
        return phylogeny_df.with_columns(
            alifestd_make_ancestor_list_col_polars(
                phylogeny_df["id"],
                phylogeny_df["ancestor_id"],
                root_ancestor_token=root_ancestor_token,
            ).alias("ancestor_list")
        )

    return phylogeny_df



import typing
from unittest import mock

import numpy as np
import tqdm

class Records: ...

def records_to_dict(records: Records) -> dict[str, np.ndarray]: ...
def collapse_dropped_unifurcations(records: Records) -> Records: ...
def extend_tree_searchtable_cpp_from_exploded(
    records: Records,
    data_ids: np.ndarray,
    num_strata_depositeds: np.ndarray,
    ranks: np.ndarray,
    differentiae: np.ndarray,
    tqdm_progress_bar: typing.Union[typing.Type[tqdm.tqdm], mock.Mock],
) -> None: ...
def build_tree_searchtable_cpp_from_exploded(
    data_ids: np.ndarray,
    num_strata_depositeds: np.ndarray,
    ranks: np.ndarray,
    differentiae: np.ndarray,
    tqdm_progress_bar: typing.Union[typing.Type[tqdm.tqdm], mock.Mock],
) -> dict[str, np.ndarray]: ...
def build_tree_searchtable_cpp_from_nested(
    data_ids: typing.List[int],
    num_strata_depositeds: typing.List[int],
    ranks: typing.List[typing.List[int]],
    differentiae: typing.List[typing.List[int]],
    tqdm_progress_bar: typing.Union[typing.Type[tqdm.tqdm], mock.Mock],
) -> dict[str, np.ndarray]: ...



import typing

from .._detail import PolicySpecBase


class PolicySpec(PolicySpecBase):
    """Contains all policy parameters, if any."""

    _retention_probability: float

    def __init__(
        self: "PolicySpec",
        *,
        retention_probability: float = 0.5,
    ) -> None:
        """Construct the policy spec.

        Parameters
        ----------
        retention_probability : float
            The proability with which strata are retained permanently.
        """
        self._retention_probability = retention_probability

    def __eq__(self: "PolicySpec", other: typing.Any) -> bool:
        return isinstance(other, self.__class__)

    def __repr__(self: "PolicySpec") -> str:
        return f"""{
            self.GetAlgoIdentifier()
        }.{
            PolicySpec.__qualname__
        }(retention_probability={
            self._retention_probability
        })"""

    def __str__(self: "PolicySpec") -> str:
        return self.GetAlgoTitle()

    def GetEvalCtor(self: "PolicySpec") -> str:
        return f"""hstrat.{self!r}"""

    def GetRetentionProbability(self: "PolicySpec") -> int:
        return self._retention_probability

    @staticmethod
    def GetAlgoIdentifier() -> str:
        """Get programatic name for underlying retention algorithm."""
        return __package__.split(".")[-1]

    @staticmethod
    def GetAlgoTitle() -> str:
        """Get human-readable name for underlying retention algorithm."""
        return "Stochastic Stratum Retention Algorithm"



import typing

from .._auxiliary_lib import HereditaryStratigraphicArtifact
from ._calc_min_implausible_spurious_consecutive_differentia_collisions_between import (
    calc_min_implausible_spurious_consecutive_differentia_collisions_between,
)
from ._calc_rank_of_first_retained_disparity_between import (
    calc_rank_of_first_retained_disparity_between,
)


def calc_definitive_max_rank_of_first_retained_disparity_between(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
) -> typing.Optional[int]:
    """Determine hard, exclusive upper bound on MRCA generation.

    At most, how many depositions elapsed along the columns' lines of
    descent before the first mismatching strata at the same rank between
    self and second?

    Returns
    -------
    int, optional
        The number of depositions elapsed or None if no disparity (i.e.,
        both columns have same number of strata deposited and the most
        recent stratum is common between first and second).

    Notes
    -----
    If no mismatching strata are found but first and second have different
    numbers of strata deposited, this method returns one greater than the
    lesser of the columns' deposition counts.
    """
    confidence_level = 0.49
    assert (
        first.GetStratumDifferentiaBitWidth()
        == second.GetStratumDifferentiaBitWidth()
    )
    assert (
        calc_min_implausible_spurious_consecutive_differentia_collisions_between(
            first,
            second,
            significance_level=1.0 - confidence_level,
        )
        == 1
    )
    return calc_rank_of_first_retained_disparity_between(
        first,
        second,
        confidence_level=confidence_level,
    )



import os

import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_aggregate_phylogenies,
    alifestd_count_leaf_nodes,
    alifestd_downsample_tips_asexual,
    alifestd_prune_extinct_lineages_asexual,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
                pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            ]
        ),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize("n_downsample", [1, 5, 10, 100000000])
@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("seed", [1, 42])
def test_alifestd_downsample_tips_asexual(
    phylogeny_df, n_downsample, mutate, seed
):
    original_df = phylogeny_df.copy()

    result_df = alifestd_downsample_tips_asexual(
        phylogeny_df, n_downsample, mutate, seed
    )

    assert len(result_df) <= len(original_df)
    assert "extant" not in result_df.columns

    if not mutate:
        pd.testing.assert_frame_equal(phylogeny_df, original_df)

    assert all(result_df["id"].isin(original_df["id"]))
    assert alifestd_count_leaf_nodes(result_df) == min(
        alifestd_count_leaf_nodes(original_df), n_downsample
    )


@pytest.mark.parametrize("n_downsample", [0, 1])
def test_alifestd_downsample_tips_asexual_with_zero_tips(n_downsample):
    phylogeny_df = pd.DataFrame({"id": [], "parent_id": [], "ancestor_id": []})

    result_df = alifestd_downsample_tips_asexual(phylogeny_df, n_downsample)

    assert result_df.empty


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
                pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            ]
        ),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize("n_downsample", [1, 5, 10])
@pytest.mark.parametrize("seed", [1, 42])
def test_prune_tips_vs_downsample(phylogeny_df, n_downsample, seed):
    original_df = phylogeny_df.copy()

    downsampled_df = alifestd_downsample_tips_asexual(
        phylogeny_df, n_downsample, seed=seed
    )

    tips_to_keep = downsampled_df["id"].tolist()
    original_df["extant"] = original_df["id"].isin(tips_to_keep)

    pruned_df = alifestd_prune_extinct_lineages_asexual(
        original_df, mutate=True
    ).drop(columns=["extant"])

    pd.testing.assert_frame_equal(pruned_df, downsampled_df)



import anytree

from ._anytree_has_sibling import anytree_has_sibling
from ._anytree_iterative_deepcopy import anytree_iterative_deepcopy


def anytree_peel_sibling_to_cousin(node: anytree.Node) -> None:
    """Peels a node from its sibling group and moves it to a new cousin
    position in the tree.

    Parameters
    ----------
    node : anytree.Node
        The node to be peeled.

    Raises
    ------
    AssertionError
        If the node is a leaf node or has no sibling or no grandparent.

    Notes
    -----
    This function clones the parent node of the input node and attaches it to
    the input node's grandparent node. The input node is then grafted away from
    its original parent and onto the newly cloned parent node as a child. The
    original parent node is left in place with any remaining children.

    Examples
    --------
    Create a tree with 3 levels of nodes, and peel a sibling node:

    >>> import anytree
    >>> from hstrat._auxiliary_lib import anytree_peel_sibling_to_cousin
    >>> a = anytree.Node("A")
    >>> b = anytree.Node("B", parent=a)
    >>> c = anytree.Node("C", parent=a)
    >>> d = anytree.Node("D", parent=b)
    >>> e = anytree.Node("E", parent=b)
    >>> f = anytree.Node("F", parent=e)
    >>> g = anytree.Node("G", parent=e)
    >>> print(anytree.RenderTree(a))
    Node('/A')
    ├── Node('/A/B')
    │   ├── Node('/A/B/D')
    │   └── Node('/A/B/E')
    │       ├── Node('/A/B/E/F')
    │       └── Node('/A/B/E/G')
    └── Node('/A/C')
    >>> anytree_peel_sibling_to_cousin(e)
    >>> print(anytree.RenderTree(a))
    Node('/A')
    ├── Node('/A/C')
    ├── Node('/A/B')
    │   └── Node('/A/B/D')
    └── Node('/A/B')
        └── Node('/A/B/E')
            ├── Node('/A/B/E/F')
            └── Node('/A/B/E/G')
    """
    assert node.parent is not None
    assert node.parent.parent is not None
    assert anytree_has_sibling(node)

    parent = node.parent

    peeled_parent = anytree_iterative_deepcopy(parent)
    peeled_parent.parent = parent.parent

    parent.children = filter(lambda x: x is not node, parent.children)

    assert node.parent is None
    peeled_parent.children = [node]



import math
import typing

from ..._detail import PolicyCouplerBase
from .._PolicySpec import PolicySpec


class CalcNumStrataRetainedUpperBound:
    """Functor to provide member function implementation in Policy class."""

    def __init__(
        self: "CalcNumStrataRetainedUpperBound",
        policy_spec: typing.Optional[PolicySpec] = None,
    ) -> None:
        pass

    def __eq__(
        self: "CalcNumStrataRetainedUpperBound",
        other: typing.Any,
    ) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "CalcNumStrataRetainedUpperBound",
        policy: PolicyCouplerBase,
        num_strata_deposited: int,
    ) -> int:
        """At most how many strata are retained after n deposited? Inclusive."""
        spec = policy.GetSpec()
        resolution = spec.GetRecencyProportionalResolution()

        res = (
            int(
                (resolution + 1) * math.log2(num_strata_deposited - 1)
                - sum(math.log2(r + 1) for r in range(resolution))
                + 1
                + (
                    resolution == 0
                )  # <<< patches failed 0-resolution test cases
            )
            if num_strata_deposited > resolution + 1
            else num_strata_deposited
        )

        return min(res, num_strata_deposited)



import numpy as np
import pandas as pd

from ._alifestd_has_compact_ids import alifestd_has_compact_ids
from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_parse_ancestor_ids import alifestd_parse_ancestor_ids
from ._jit import jit
from ._jit_numpy_bool_t import jit_numpy_bool_t


@jit(nopython=True)
def _is_topologically_sorted_contiguous(ancestor_ids: np.ndarray) -> bool:
    for id_, ancestor_id in enumerate(ancestor_ids):
        if id_ < ancestor_id:
            return False
    return True


@jit(nopython=True)
def _is_topologically_sorted_compact(
    ids: np.ndarray, ancestor_ids: np.ndarray
) -> bool:
    seen_ancestor_ids = np.zeros(len(ids), dtype=jit_numpy_bool_t)
    for id_, ancestor_id in zip(ids, ancestor_ids):
        if seen_ancestor_ids[id_]:
            return False
        seen_ancestor_ids[ancestor_id] = True
    return True


@jit(nopython=True)
def _is_topologically_sorted(
    ids: np.ndarray, ancestor_ids: np.ndarray
) -> bool:
    seen_ancestor_ids = set()
    for id_, ancestor_id in zip(ids, ancestor_ids):
        if id_ in seen_ancestor_ids:
            return False
        seen_ancestor_ids.add(ancestor_id)
    return True


def alifestd_is_topologically_sorted(phylogeny_df: pd.DataFrame) -> bool:
    """Are all organisms listed after members of their `ancestor_list`?

    Input dataframe is not mutated by this operation.
    """

    if "ancestor_id" in phylogeny_df:
        if alifestd_has_contiguous_ids(phylogeny_df):
            return _is_topologically_sorted_contiguous(
                phylogeny_df["ancestor_id"].to_numpy()
            )
        elif alifestd_has_compact_ids(phylogeny_df):
            return _is_topologically_sorted_compact(
                phylogeny_df["id"].to_numpy(),
                phylogeny_df["ancestor_id"].to_numpy(),
            )
        else:
            return _is_topologically_sorted(
                phylogeny_df["id"].to_numpy(),
                phylogeny_df["ancestor_id"].to_numpy(),
            )

    phylogeny_df = phylogeny_df.set_index("id", drop=False)

    for pos, (_idx, row) in enumerate(phylogeny_df.iterrows()):
        for ancestor_id in alifestd_parse_ancestor_ids(row["ancestor_list"]):
            if phylogeny_df.index.get_loc(ancestor_id) >= pos:
                return False

    return True



import os

import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_convert_root_ancestor_token,
    alifestd_make_ancestor_list_col,
    alifestd_to_working_format,
    alifestd_try_add_ancestor_id_col,
    alifestd_try_add_ancestor_list_col,
    alifestd_validate,
)

from ._impl import enforce_identical_polars_result

assets_path = os.path.join(os.path.dirname(__file__), "assets")


alifestd_try_add_ancestor_list_col_ = enforce_identical_polars_result(
    alifestd_try_add_ancestor_list_col,
)


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        alifestd_to_working_format,
        lambda x: x,
    ],
)
def test_alifestd_try_add_ancestor_id_col_asexual(phylogeny_df, apply):
    phylogeny_df = alifestd_try_add_ancestor_id_col(
        apply(phylogeny_df),
    )
    phylogeny_df["ancestor_list"] = alifestd_convert_root_ancestor_token(
        phylogeny_df["ancestor_list"],
        root_ancestor_token="None",
    )
    phylogeny_df_ = phylogeny_df.copy()

    assert alifestd_validate(phylogeny_df)
    assert (
        phylogeny_df["ancestor_list"]
        == alifestd_make_ancestor_list_col(
            phylogeny_df["id"],
            phylogeny_df["ancestor_id"],
            root_ancestor_token="None",
        )
    ).all(), phylogeny_df
    pd.testing.assert_frame_equal(phylogeny_df_, phylogeny_df)

    res_df = alifestd_try_add_ancestor_list_col_(
        phylogeny_df,
        root_ancestor_token="None",
    )
    assert (
        res_df["ancestor_list"]
        == alifestd_make_ancestor_list_col(
            res_df["id"],
            res_df["ancestor_id"],
            root_ancestor_token="None",
        )
    ).all()
    pd.testing.assert_frame_equal(phylogeny_df_, phylogeny_df)

    res_df = alifestd_try_add_ancestor_list_col_(
        phylogeny_df,
        root_ancestor_token="None",
    )
    assert (
        res_df["ancestor_list"]
        == alifestd_make_ancestor_list_col(
            res_df["id"],
            res_df["ancestor_id"],
            root_ancestor_token="None",
        )
    ).all()

    res_df = alifestd_try_add_ancestor_list_col_(
        phylogeny_df,
        root_ancestor_token="None",
    )
    assert (
        res_df["ancestor_list"]
        == alifestd_make_ancestor_list_col(
            res_df["id"],
            res_df["ancestor_id"],
            root_ancestor_token="None",
        )
    ).all()

    res_df = res_df.copy()
    res_df.drop(columns="ancestor_list", inplace=True)
    assert "ancestor_list" not in res_df
    res_df = alifestd_try_add_ancestor_list_col_(
        res_df,
        root_ancestor_token="None",
    )
    assert (
        res_df["ancestor_list"]
        == alifestd_make_ancestor_list_col(
            res_df["id"],
            res_df["ancestor_id"],
            root_ancestor_token="None",
        )
    ).all()

    phylogeny_df.drop(columns="ancestor_id", inplace=True)
    assert "ancestor_id" not in phylogeny_df
    res_df = alifestd_try_add_ancestor_list_col_(phylogeny_df)
    pd.testing.assert_frame_equal(res_df, phylogeny_df)

    phylogeny_df.drop(columns="ancestor_list", inplace=True)
    assert "ancestor_list" not in phylogeny_df
    res_df = alifestd_try_add_ancestor_list_col_(phylogeny_df)
    pd.testing.assert_frame_equal(res_df, phylogeny_df)



import pandas as pd

from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_is_topologically_sorted import alifestd_is_topologically_sorted
from ._alifestd_topological_sort import alifestd_topological_sort
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col


def alifestd_mark_num_children_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """Add column `num_children`, counting for each node the number of nodes it
    is parent to.

    A topological sort will be applied if `phylogeny_df` is not topologically
    sorted. Reindexing may be applied.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    if not alifestd_is_topologically_sorted(phylogeny_df):
        phylogeny_df = alifestd_topological_sort(phylogeny_df, mutate=True)

    if alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df.reset_index(drop=True, inplace=True)
    else:
        phylogeny_df.index = phylogeny_df["id"]

    phylogeny_df["num_children"] = 0

    for idx in reversed(phylogeny_df.index):
        ancestor_id = phylogeny_df.loc[idx, "ancestor_id"]
        if ancestor_id == idx:
            continue  # handle genesis cases

        phylogeny_df.loc[ancestor_id, "num_children"] += 1

    return phylogeny_df



from ..._impl import CalcNumStrataRetainedUpperBoundWorstCase


# must inherit rather than assign due to failure of attribute lookup when
# pickling otherwise
class CalcNumStrataRetainedUpperBound(
    CalcNumStrataRetainedUpperBoundWorstCase
):
    pass



import typing


def deep_listify(maybe_iterable: typing.Iterable) -> typing.Any:
    """Recursively convert an iterable into a list, converting any nested
    iterables recusrively.

    Will not convert strings.

    Parameters
    ----------
    maybe_iterable
        The object to be converted.

    Returns
    -------
        A nested list containing the elements of `maybe_iterable` if `maybe_iterable` was iterable, else `maybe_iterable`.


    Examples
    --------
    >>> deep_listify(range(4))
    [0, 1, 2, 3]

    >>> deep_listify(4)
    4

    >>> deep_listify([(1, 2), 3, 4])
    [[1, 2], 3, 4]

    >>> deep_listify([1, [2, (3, 4)], 5])
    [1, [2, [3, 4]], 5]

    >>> deep_listify(iter([1, range(2), 3]))
    [1, [0, 1], 3]
    """
    try:
        if isinstance(maybe_iterable, str):
            raise TypeError
        else:
            return [*map(deep_listify, maybe_iterable)]
    except TypeError:
        return maybe_iterable



from hstrat._auxiliary_lib import coshuffled


def test_coshuffle_conserve_elements():
    a = [1, 2, 3, 4, 5]
    b = [5, 6, 7, 8, 9]
    c = [9, 10, 11, 12, 13]
    a_, b_, c_ = coshuffled(a, b, c)

    assert a != a_
    assert b != b_
    assert c != c_

    assert sorted(a) == sorted(a_)
    assert sorted(b) == sorted(b_)
    assert sorted(c) == sorted(c_)


def test_coshuffle_keep_relative_order():
    a = [1, 2, 3, 4]
    b = [5, 6, 7, 8]
    assert all(y - x == 4 for x, y in zip(*coshuffled(a, b)))


def test_coshuffle_singleton():
    assert ([1], [2], [3]) == coshuffled([1], [2], [3])


def test_coshuffle_empty():
    assert ([], [], []) == coshuffled([], [], [])


def test_coshuffle_singleton2():
    a = [1]
    assert coshuffled(a) == ([1],)


def test_coshuffle_empty2():
    a = []
    assert coshuffled(a) == ([],)


def test_coshuffle_empty3():
    assert coshuffled() == tuple()



import pytest

from hstrat.hstrat import recency_proportional_resolution_curbed_algo


@pytest.mark.parametrize(
    "size_curb",
    [
        2,
        3,
        7,
        42,
        97,
        100,
    ],
)
def test_eq(size_curb):
    policy = recency_proportional_resolution_curbed_algo.Policy(size_curb)
    spec = policy.GetSpec()
    instance = recency_proportional_resolution_curbed_algo.IterRetainedRanks(
        spec
    )

    assert instance == instance
    assert (
        instance
        == recency_proportional_resolution_curbed_algo.IterRetainedRanks(spec)
    )
    assert instance is not None



import anytree

from hstrat._auxiliary_lib import anytree_calc_leaf_counts


def test_anytree_calc_leaf_counts_single_node():
    # Test for a tree with a single node
    root = anytree.Node("A")
    expected_res = {id(root): 1}
    assert anytree_calc_leaf_counts(root) == expected_res


def test_anytree_calc_leaf_counts_multiple_nodes():
    # Example tree structure
    # Node('/A')
    # ├── Node('/A/B')
    # │   ├── Node('/A/B/D')
    # │   └── Node('/A/B/E')
    # └── Node('/A/C')
    #     ├── Node('/A/C/F')
    #     └── Node('/A/C/G')
    #         └── Node('/A/C/G/H')
    root = anytree.Node("A")
    b = anytree.Node("B", parent=root)
    c = anytree.Node("C", parent=root)
    d = anytree.Node("D", parent=b)
    e = anytree.Node("E", parent=b)
    f = anytree.Node("F", parent=c)
    g = anytree.Node("G", parent=c)
    h = anytree.Node("H", parent=g)

    # Test for a tree with multiple nodes
    expected_res = {
        id(root): 4,
        id(b): 2,
        id(c): 2,
        id(d): 1,
        id(e): 1,
        id(f): 1,
        id(g): 1,
        id(h): 1,
    }
    assert anytree_calc_leaf_counts(root) == expected_res



import numpy as np
import pandas as pd

from hstrat._auxiliary_lib import as_nullable_type


def test_as_nullable_type_integer():
    s_int = pd.Series([1, 2, 3, 4], dtype="int16")
    s_int_nullable = as_nullable_type(s_int)
    print(s_int_nullable)
    assert s_int_nullable.dtype == "Int16"
    assert all(s_int_nullable == s_int)


def test_as_nullable_type_unsigned_integer():
    s_int = pd.Series([1, 2, 3, 4], dtype="uint8")
    s_int_nullable = as_nullable_type(s_int)
    print(s_int_nullable)
    assert s_int_nullable.dtype == "UInt8"
    assert all(s_int_nullable == s_int)


def test_as_nullable_type_Integer():
    s_int = pd.Series([1, 2, 3, 4, pd.NA], dtype="Int64")
    s_int_nullable = as_nullable_type(s_int)
    print(s_int_nullable)
    assert s_int_nullable.dtype == "Int64"
    pd.testing.assert_series_equal(s_int_nullable, s_int)


def test_as_nullable_type_float():
    s_float = pd.Series([1.0, 2.0, 3.0, np.nan, 4.0])
    s_float_nullable = as_nullable_type(s_float)
    assert s_float_nullable.dtype == "float64"
    pd.testing.assert_series_equal(s_float_nullable, s_float)


def test_as_nullable_type_bool():
    s_bool = pd.Series([True, False, True])
    s_bool_nullable = as_nullable_type(s_bool)
    assert s_bool_nullable.dtype == "boolean"
    assert all(s_bool_nullable == s_bool)


def test_as_nullable_type_boolean():
    s_bool = pd.Series([True, False, True], dtype="boolean")
    s_bool_nullable = as_nullable_type(s_bool)
    assert s_bool_nullable.dtype == "boolean"
    pd.testing.assert_series_equal(s_bool_nullable, s_bool)


def test_as_nullable_type_mixed_type():
    s_mixed = pd.Series([1, "foo", True])
    s_mixed_nullable = as_nullable_type(s_mixed)
    assert s_mixed_nullable.dtype == "object"
    pd.testing.assert_series_equal(s_mixed_nullable, s_mixed)



from ._HereditaryStratumOrderedStoreDict import (
    HereditaryStratumOrderedStoreDict,
)
from ._HereditaryStratumOrderedStoreList import (
    HereditaryStratumOrderedStoreList,
)
from ._HereditaryStratumOrderedStoreTree import (
    HereditaryStratumOrderedStoreTree,
)
from ._provided_stratum_ordered_stores import provided_stratum_ordered_stores

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "HereditaryStratumOrderedStoreDict",
    "HereditaryStratumOrderedStoreList",
    "HereditaryStratumOrderedStoreTree",
    "provided_stratum_ordered_stores",
]



import typing

from .._detail import PolicySpecBase


class PolicySpec(PolicySpecBase):
    """Contains all policy parameters, if any."""

    _fixed_resolution: int

    def __init__(
        self: "PolicySpec",
        fixed_resolution: int,
    ):
        """Construct the policy spec.

        Parameters
        ----------
        fixed_resolution : int
            The rank interval strata should be retained at. The uncertainty of
            MRCA estimates provided under the fixed resolution policy will
            always be strictly less than this cap.
        """
        assert fixed_resolution > 0
        self._fixed_resolution = fixed_resolution

    def __eq__(self: "PolicySpec", other: typing.Any) -> bool:
        return isinstance(other, self.__class__) and (
            self._fixed_resolution,
        ) == (other._fixed_resolution,)

    def __repr__(self: "PolicySpec") -> str:
        return f"""{
            self.GetAlgoIdentifier()
        }.{
            PolicySpec.__qualname__
        }(fixed_resolution={
            self._fixed_resolution
        })"""

    def __str__(self: "PolicySpec") -> str:
        return f"""{
            self.GetAlgoTitle()
        } (resolution: {
            self._fixed_resolution
        })"""

    def GetEvalCtor(self: "PolicySpec") -> str:
        return f"hstrat.{self!r}"

    def GetFixedResolution(self: "PolicySpec") -> int:
        return self._fixed_resolution

    @staticmethod
    def GetAlgoIdentifier() -> str:
        """Get programatic name for underlying retention algorithm."""
        return __package__.split(".")[-1]

    @staticmethod
    def GetAlgoTitle() -> str:
        """Get human-readable name for underlying retention algorithm."""
        return "Fixed Resolution Stratum Retention Algorithm"



import typing

from ..._detail import PolicyCouplerBase
from .._PolicySpec import PolicySpec
from .._impl import calc_provided_uncertainty


class CalcMrcaUncertaintyAbsExact:
    """Functor to provide member function implementation in Policy class."""

    def __init__(
        self: "CalcMrcaUncertaintyAbsExact",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(self: "CalcMrcaUncertaintyAbsExact", other: typing.Any) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "CalcMrcaUncertaintyAbsExact",
        policy: PolicyCouplerBase,
        first_num_strata_deposited: int,
        second_num_strata_deposited: int,
        actual_rank_of_mrca: int,
    ) -> int:
        """Exactly how much uncertainty to estimate rank of MRCA?"""
        # rectify negative-indexed actual_rank_of_mrca
        if actual_rank_of_mrca is not None and actual_rank_of_mrca < 0:
            least_last_rank = min(
                first_num_strata_deposited - 1,
                second_num_strata_deposited - 1,
            )
            actual_rank_of_mrca += least_last_rank
            assert actual_rank_of_mrca >= 0

        spec = policy.GetSpec()
        resolution = spec.GetRecencyProportionalResolution()

        least_num_strata_deposited = min(
            first_num_strata_deposited,
            second_num_strata_deposited,
        )
        least_last_rank = least_num_strata_deposited - 1
        # mrca at very last rank
        if least_last_rank - actual_rank_of_mrca <= resolution:
            return 0

        provided_uncertainty = calc_provided_uncertainty(
            resolution,
            least_last_rank,
        )
        if provided_uncertainty == 1:
            return 0

        # we could just take a simple recursive approach like this
        #
        #   if index == 0: return 0
        #   else: return provided_uncertainty + self.CalcRankAtColumnIndex(
        #       index - 1,
        #       num_strata_deposited - provided_uncertainty,
        #   )
        #
        # but as an optimization, we should take as many steps as possible at
        # the current uncertainty interval size stage before recursing down to
        # the next half-as-small uncertainty interval size

        # calculate the greatest rank at which that interval is viable
        # i.e., where max_uncertainty == provided_uncertainty... we must solve
        #
        #   provided_uncertainty = least_num_strata // (resolution + 1)
        #
        # and
        #
        #   least_num_strata = num_strata_deposited - greatest_viable_rank

        # -1 due to *lack* of an in-progress deposition
        greatest_viable_rank = least_last_rank - provided_uncertainty * (
            resolution + 1
        )

        # calculate how many steps at provided_uncertainty interval we can take
        # +provided_uncertainty because greatest viable rank is the *last*
        # position we can take a step in our interval from
        num_interval_steps = (
            greatest_viable_rank + provided_uncertainty
        ) // provided_uncertainty
        cutoff_rank = num_interval_steps * provided_uncertainty

        if actual_rank_of_mrca < cutoff_rank:
            if actual_rank_of_mrca >= (
                least_last_rank - least_last_rank % provided_uncertainty
            ):
                return (least_last_rank - 1) % provided_uncertainty
            else:
                return provided_uncertainty - 1
        else:
            assert cutoff_rank
            return self(
                policy=policy,
                first_num_strata_deposited=least_num_strata_deposited
                - cutoff_rank,
                second_num_strata_deposited=least_num_strata_deposited
                - cutoff_rank,
                actual_rank_of_mrca=actual_rank_of_mrca - cutoff_rank,
            )



import unittest

import contexttimer as ctt
import numpy as np
import pandas as pd

from hstrat._auxiliary_lib import indices_of_unique


class TestIndicesOfUnique(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test_empty(self):
        assert len(indices_of_unique(np.array([]))) == 0

    def test_singleton(self):
        assert all(indices_of_unique(np.array([-1])) == np.array([0]))
        assert all(indices_of_unique(np.array([0])) == np.array([0]))
        assert all(indices_of_unique(np.array([1])) == np.array([0]))

    def test_pair(self):
        assert all(indices_of_unique(np.array([-1, 1])) == np.array([0, 1]))
        assert all(indices_of_unique(np.array([-1, -1])) == np.array([0]))
        assert all(indices_of_unique(np.array([0, 0])) == np.array([0]))
        assert all(indices_of_unique(np.array([0, 1])) == np.array([0, 1]))

    def test_triple(self):
        assert all(
            indices_of_unique(np.array([0, -1, 1])) == np.array([0, 1, 2])
        )
        assert all(
            indices_of_unique(np.array([-1, 1, -1])) == np.array([0, 1])
        )
        assert all(indices_of_unique(np.array([1, 0, 0])) == np.array([0, 1]))
        assert all(indices_of_unique(np.array([0, 0, 1])) == np.array([0, 2]))
        assert all(indices_of_unique(np.array([0, 0, 0])) == np.array([0]))

    def test_benchmark(self):
        d1 = np.arange(1000000)
        np.random.shuffle(d1)
        d2 = np.repeat(np.arange(1000), 1000)
        np.random.shuffle(d2)
        d3 = np.repeat(np.arange(1000), 1000)

        with ctt.Timer(factor=1000) as t1_hstrat:
            indices_of_unique(d1)

        with ctt.Timer(factor=1000) as t2_hstrat:
            indices_of_unique(d2)

        with ctt.Timer(factor=1000) as t3_hstrat:
            indices_of_unique(d3)

        with ctt.Timer(factor=1000) as t1_numpy:
            np.unique(d1, return_index=True)

        with ctt.Timer(factor=1000) as t2_numpy:
            np.unique(d2, return_index=True)

        with ctt.Timer(factor=1000) as t3_numpy:
            np.unique(d3, return_index=True)

        with ctt.Timer(factor=1000) as t1_pandas:
            pd.unique(d1)

        with ctt.Timer(factor=1000) as t2_pandas:
            pd.unique(d2)

        with ctt.Timer(factor=1000) as t3_pandas:
            pd.unique(d3)

        print()
        print(
            f"t1_hstrat.elapsed={t1_hstrat.elapsed} "
            f"t2_hstrat.elapsed={t2_hstrat.elapsed} "
            f"t3_hstrat.elapsed={t3_hstrat.elapsed} "
        )
        print(
            f"t1_numpy.elapsed={t1_numpy.elapsed} "
            f"t2_numpy.elapsed={t2_numpy.elapsed} "
            f"t3_numpy.elapsed={t3_numpy.elapsed} "
        )
        print(
            f"t1_pandas.elapsed={t1_pandas.elapsed} "
            f"t2_pandas.elapsed={t2_pandas.elapsed} "
            f"t3_pandas.elapsed={t3_pandas.elapsed} "
        )


if __name__ == "__main__":
    unittest.main()



from functools import reduce
import operator
import typing

import opytional as opyt


def find_bounds(
    query: typing.Any,
    iterable: typing.Iterable[typing.Any],
    filter_below: typing.Callable = operator.lt,
    filter_above: typing.Callable = operator.gt,
    key: typing.Callable = lambda x: x,
    initializer: typing.Tuple[typing.Any, typing.Any] = (None, None),
) -> typing.Tuple[typing.Optional[typing.Any], typing.Optional[typing.Any]]:
    def operation(accumulation, element):
        cur_below, cur_above = accumulation

        element = key(element)
        maybe_below = element if filter_below(element, query) else None
        maybe_above = element if filter_above(element, query) else None

        res_below = opyt.apply_if_or_value(
            cur_below,
            lambda x: opyt.apply_if_or_value(
                maybe_below,
                lambda y: max(x, y),
                x,
            ),
            maybe_below,
        )
        res_above = opyt.apply_if_or_value(
            cur_above,
            lambda x: opyt.apply_if_or_value(
                maybe_above,
                lambda y: min(x, y),
                x,
            ),
            maybe_above,
        )

        return (res_below, res_above)

    return reduce(operation, iterable, initializer)



import contextlib
import typing

import opytional as opyt
import pandas as pd

from ..._auxiliary_lib import (
    HereditaryStratigraphicArtifact,
    RngStateContext,
    alifestd_collapse_unifurcations,
    alifestd_make_empty,
    anytree_tree_to_alife_dataframe,
    flag_last,
)
from ...juxtaposition import calc_probability_differentia_collision_between
from ._impl import TrieInnerNode, build_trie_from_artifacts


def _finalize_trie(trie: TrieInnerNode) -> pd.DataFrame:
    "Convert to alifestd dataframe and collapse unifurcations."
    return alifestd_collapse_unifurcations(
        anytree_tree_to_alife_dataframe(trie), mutate=True
    )


def _build_tree_trie_ensemble(
    population: typing.Sequence[HereditaryStratigraphicArtifact],
    trie_postprocessors: typing.Iterable[typing.Callable],
    taxon_labels: typing.Optional[typing.Iterable],
    force_common_ancestry: bool,
    progress_wrap: typing.Callable,
) -> typing.List[pd.DataFrame]:
    """Implementation detail for build_tree_trie_ensemble.

    See `build_tree_trie_ensemble` for parameter descriptions.
    """
    # for simplicity, return early for this special case
    if len(population) == 0:
        return alifestd_make_empty()

    root = build_trie_from_artifacts(
        population=population,
        taxon_labels=taxon_labels,
        progress_wrap=progress_wrap,
    )
    if not force_common_ancestry:
        try:
            (root,) = root.children
            root.parent = None
        except ValueError:
            raise ValueError(
                "Reconstruction resulted in multiple independent trees, "
                "due to artifacts definitively sharing no common ancestor. "
                "Consider setting force_common_ancestry=True.",
            )

    p_differentia_collision = calc_probability_differentia_collision_between(
        population[0], population[0]
    )

    res = []
    for is_last, postprocessor in flag_last(trie_postprocessors):
        res.append(
            _finalize_trie(
                postprocessor(
                    root,
                    p_differentia_collision=p_differentia_collision,
                    mutate=is_last,
                    progress_wrap=progress_wrap,
                ),
            ),
        )

    return res


def build_tree_trie_ensemble(
    population: typing.Sequence[HereditaryStratigraphicArtifact],
    trie_postprocessors: typing.Sequence[typing.Callable],
    taxon_labels: typing.Optional[typing.Iterable] = None,
    force_common_ancestry: bool = False,
    progress_wrap: typing.Callable = lambda x: x,
    seed: typing.Optional[int] = 1,
) -> typing.List[pd.DataFrame]:
    """Estimate the phylogenetic history among hereditary stratigraphic
    columns by building a trie (a.k.a. prefix tree) of the differentia
    sequences of hereditary stratigraphic artifacts within a population.

    Returns phylogeny reconstruction outcomes from alternate postprocessing
    schemes applied between trie contruction and conversion to an alife
    standard data frame, including ancestor taxon origin time estimation.
    Because the underlying pre-postprocess trie is only constructed once, this
    method allows for efficient comparison of potprocessing schemes.

    Unless comparing alternate postprocessing schemes or applying a custom
    postprocessing option, end users should likely prefer `build_tree_trie`.
    This interface applies a single postprocess, set to a generally-appropriate
    default with a few other curated postprocesses specifiable by optional
    argument.

    Parameters
    ----------
    population: Sequence[HereditaryStratigraphicArtifact]
        Hereditary stratigraphic columns corresponding to extant population
        members.

        Each member of population will correspond to a unique leaf node in the
        reconstructed tree.
    trie_postprocessors: Iterable[Callable]
        Tree postprocess functors.

        Must take `trie` of type `TrieInnerNode`, `p_differentia_collision` of
        type `float`, `mutate` of type `bool`, and `progress_wrap` of type
        `Callable` params. Must returned postprocessed trie (type
        `TrieInnerNode`). Several

        Each postprocess will be called indpendently to produce a returned
        postprocessed variant. Use `CompoundTriePostprocessor` to chain
        postprocess functors that should be applied in succession.
    taxon_labels: Optional[Iterable], optional
        How should leaf nodes representing extant hereditary stratigraphic
        columns be named?

        Label order should correspond to the order of corresponding hereditary
        stratigraphic columns within `population`. If None, taxons will be
        named according to their numerical index.
    force_common_ancestry: bool, default False
        How should columns that definitively share no common ancestry be
        handled?

        If set to True, treat columns with no common ancestry as if they
        shared a common ancestor immediately before the genesis of the
        lineages. If set to False, columns within `population` that
        definitively do not share common ancestry will raise a ValueError.
    progress_wrap : Callable, default identity function
        Wrapper applied around generation iterator and row generator for final
        phylogeny compilation process.

        Pass tqdm or equivalent to display progress bars.
    seed: int, default 1
        Controls tiebreaking decisions in the algorithm.

        Pass an int for reproducible output across multiple function calls. The
        default value, 1, ensures reproducible output. Pass None to use
        existing RNG context directly.

    Returns
    -------
    typing.List[pd.DataFrame]
        Reconstructed phylogenetic trees with each postprocessor applied, in
        alife standard format.
    """
    with opyt.apply_if_or_value(
        seed,
        RngStateContext,
        contextlib.nullcontext(),
    ):
        return _build_tree_trie_ensemble(
            population=population,
            trie_postprocessors=trie_postprocessors,
            taxon_labels=taxon_labels,
            force_common_ancestry=force_common_ancestry,
            progress_wrap=progress_wrap,
        )



from ._IterRetainedRanks_ import IterRetainedRanksProduction


# must inherit rather than assign due to failure of attribute lookup when
# pickling otherwise
class IterRetainedRanks(IterRetainedRanksProduction):
    pass



import os

import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_is_asexual,
    alifestd_to_working_format,
    alifestd_try_add_ancestor_id_col,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
        ),
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        alifestd_to_working_format,
        alifestd_try_add_ancestor_id_col,
        lambda x: x,
    ],
)
def test_alifestd_is_asexual_false(phylogeny_df, apply):
    phylogeny_df = apply(phylogeny_df)
    phylogeny_df_ = phylogeny_df.copy()
    assert not alifestd_is_asexual(phylogeny_df)
    assert phylogeny_df.equals(phylogeny_df_)


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        alifestd_to_working_format,
        alifestd_try_add_ancestor_id_col,
        lambda x: x,
    ],
)
def test_alifestd_is_asexual_true(phylogeny_df, apply):
    phylogeny_df = apply(phylogeny_df)
    phylogeny_df_ = phylogeny_df.copy()
    assert alifestd_is_asexual(phylogeny_df)
    assert phylogeny_df.equals(phylogeny_df_)



import itertools as it
import typing
import warnings

from ..._auxiliary_lib import HereditaryStratigraphicArtifact


def does_definitively_share_no_common_ancestor(
    population: typing.Iterable[HereditaryStratigraphicArtifact],
) -> typing.Optional[bool]:
    """Could the population possibly share a common ancestor?

    Note that stratum rention policies are strictly required to permanently
    retain the most ancient stratum.

    Returns None if population is empty or singleton.

    See Also
    --------
    does_share_any_common_ancestor:
        Can we conclude with confidence_level confidence that the population
        shares a common ancestor?
    """
    pop_tee1, pop_tee2 = it.tee(population)
    if len([*zip(pop_tee1, range(2))]) < 2:
        warnings.warn(
            "Empty or singleton population. Unable to detect if does "
            "definitively share no common ancestor."
        )
        return None

    num_unique_rank_0_differentia = len(
        {
            (
                # note: columns guaranteed never empty
                next(column.IterRetainedDifferentia()),
                column.GetStratumDifferentiaBitWidth(),
            )
            for column in pop_tee2
        }
    )

    return num_unique_rank_0_differentia > 1



from ..._auxiliary_lib import HereditaryStratigraphicArtifact
from ...juxtaposition import (
    calc_definitive_max_rank_of_first_retained_disparity_between,
)


def does_definitively_have_no_common_ancestor(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
) -> bool:
    """Does the hereditary stratigraphic record definitively prove that first
    and second could not possibly share a common ancestor?

    If the founding strata of first and second (i.e., generation 0) have
    unequal differentia, then first and second cannot possibly share common
    ancestry.

    Note equal differentia at generation 0 does not necessarily imply common
    ancestry; colliding differentia values could have been independently
    generated by chance.

    Note also that stratum rention policies are strictly required to permanently
    retain the most ancient stratum.

    See Also
    --------
    does_have_any_common_ancestor :
        Can we conclude with confidence_level confidence that first and second
        share a common ancestor?
    """
    first_disparity = (
        calc_definitive_max_rank_of_first_retained_disparity_between(
            first,
            second,
        )
    )
    return False if first_disparity is None else first_disparity == 0



import typing

from ..._auxiliary_lib import HereditaryStratigraphicArtifact
from ...juxtaposition import (
    calc_min_implausible_spurious_consecutive_differentia_collisions_between,
    get_nth_common_rank_between,
)


def calc_rank_of_earliest_detectable_mrca_between(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
    confidence_level: float = 0.95,
) -> typing.Optional[int]:
    """After what generation is common ancstry robustly detectable?

    Calculates the earliest possible rank a MRCA between first and second
    could be reliably detected at.

    Even if a true MRCA of first and second exists, if it occured earlier
    than the rank calculated here it could not be reliably detected with
    sufficient confidence after accounting for the possibility of spurious
    differentia collisions. (Although subsequent spurious differentia
    collisions after the true MRCA of first and second could lead to MRCA
    detection at such a rank.)

    Returns None if insufficient common ranks exist between first and second
    to ever conclude at the given confidence level the existance of any
    common ancestry between first and second (even if all strata at common
    ranks had equivalent differentiae).
    """
    assert (
        first.GetStratumDifferentiaBitWidth()
        == second.GetStratumDifferentiaBitWidth()
    )
    num_required_common_ranks = calc_min_implausible_spurious_consecutive_differentia_collisions_between(
        first,
        second,
        significance_level=1 - confidence_level,
    )
    # zero-indexed
    return get_nth_common_rank_between(
        first, second, num_required_common_ranks - 1
    )



import numpy as np

from hstrat._auxiliary_lib import as_compact_type


def test_as_compact_type_empty():
    arr = np.array([])
    assert np.array_equal(
        as_compact_type(arr), np.array([], dtype=np.dtype("int64"))
    )


def test_as_compact_type_int8():
    arr = np.array([1, 2, 3, 4, 5])
    assert np.array_equal(
        as_compact_type(arr), np.array([1, 2, 3, 4, 5], dtype=np.dtype("int8"))
    )


def test_as_compact_type_uint8():
    arr = np.array([255, 0, 1])
    assert np.array_equal(
        as_compact_type(arr), np.array([255, 0, 1], dtype=np.dtype("uint8"))
    )


def test_as_compact_type_int16():
    arr = np.array([-128, 0, 127])
    assert np.array_equal(
        as_compact_type(arr), np.array([-128, 0, 127], dtype=np.dtype("int8"))
    )


def test_as_compact_type_uint16():
    arr = np.array([-32768, 0, 32767])
    assert np.array_equal(
        as_compact_type(arr),
        np.array([-32768, 0, 32767], dtype=np.dtype("int16")),
    )


def test_as_compact_type_float16():
    arr = np.array([-0.5, 0, 0.5])
    assert np.array_equal(
        as_compact_type(arr),
        np.array([-0.5, 0, 0.5], dtype=np.dtype("float16")),
    )

    arr = np.array([1e-6])
    assert np.array_equal(
        as_compact_type(arr), np.array([1e-6], dtype=np.dtype("float16"))
    )

    arr = np.array([1e-20])
    assert np.array_equal(
        as_compact_type(arr), np.array([1e-20], dtype=np.dtype("float16"))
    )


def test_as_compact_type_float32():
    arr = np.array([-1e-8, 0, 1e8])
    assert np.array_equal(
        as_compact_type(arr),
        np.array([-1e-8, 0, 1e8], dtype=np.dtype("float32")),
    )


def test_as_compact_type_float64():
    arr = np.array([1e200])
    assert np.array_equal(
        as_compact_type(arr), np.array([1e200], dtype=np.dtype("float64"))
    )

    arr = np.array([1e-200, 1, 1e200])
    assert np.array_equal(
        as_compact_type(arr),
        np.array([1e-200, 1, 1e200], dtype=np.dtype("float64")),
    )



"""
Script to validate and analyze Python packages for correct use of `__all__` references
across modules and type stubs. Uses the AST module to parse source code for compliance
with lazy-loading and `__all__` reference conventions. The requirements are as follows:
1) In an __init__.py file where the __all__ of some subpackage is accessed, the type stub
    of said subpackage must declare an __all__ equivalent to the actual (evaluated) __all__.
2) Every symbol imported from the __all__ of a subpackage must be in the __all__ of the
    type stub of the main package.

Functions
---------
find_modules_with_all_references(package_path)
    Parses a the `__init__.py` in the `package_path` to find all subpackages from which
    all symbols are imported (via `__all__`).

get_dunder_all_from_stub(package_path)
    Extracts the `__all__` list from the `__init__.pyi` file in the `package_path`.

check_accurate_all_declarations()
    Walks the `hstrat` directory and makes sure that the `__all__` defined in a type stub
    of a package is the same as the `__all__` obtained by importing the package directly.
    Returns a generator of violations.

check_accurate_subpackage_star_imports()
    If a package imports `__all__` from a subpackages, makes sure that that subpackage's
    real `__all__` is a subset of the `__all__` declared in the type stub of the importing
    package. Returns a generator of violations.

check_accurate_flat_namespace()
    Checks the flat namespace `hstrat.hstrat` and assures that the `__all__` defined in its
    `__init__.pyi` is accurate, containing all symbols from flattened submodules.
"""

import ast
import importlib.util
import os
from pathlib import Path
from types import ModuleType
from typing import Iterable, List


def module_name_from_path(path: str) -> str:
    return path.replace(os.path.sep, ".")


def import_from_path(directory_path) -> ModuleType:
    """Import a module from a specified directory path."""
    module_name = module_name_from_path(directory_path)
    spec = importlib.util.spec_from_file_location(
        module_name, os.path.join(directory_path, "__init__.py")
    )
    assert (
        spec is not None and spec.loader is not None
    ), f"Module '{module_name}' was not found."
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def find_modules_with_all_references(package_path: str) -> list:
    """
    Parses the `__init__.py` file of the given package path using the
    AST module to find all imports of subpackages that reference `__all__`.

    Parameters
    ----------
    package_path : str
        The file path of the Python package to analyze.

    Returns
    -------
    list of str
        A list of module names from `submod_attrs` that reference `__all__`.

    Notes
    -----
    The function uses the `ast` module to parse the source file and identify
    assignments that contain a `lazy_attach` call with a `submod_attrs` dictionary,
    where module names are specified with an `__all__` reference.
    """
    modules_with_all = []
    tree = ast.parse((Path(package_path) / "__init__.py").read_text())
    for node in tree.body:
        if (
            isinstance(node, ast.Assign)
            and isinstance(call_node := node.value, ast.Call)
            and isinstance(call_node.func, ast.Name)
            and call_node.func.id == "lazy_attach"
        ):
            for keyword in call_node.keywords:
                if keyword.arg == "submod_attrs" and isinstance(
                    keyword.value, ast.Dict
                ):
                    submod_attrs_dict = keyword.value
                    for key_node, value_node in zip(
                        submod_attrs_dict.keys, submod_attrs_dict.values
                    ):
                        if not isinstance(key_node, ast.Constant):
                            continue
                        module_name = key_node.value
                        if (
                            isinstance(value_node, ast.Attribute)
                            and value_node.attr == "__all__"
                            and isinstance(value_node.value, ast.Name)
                            and module_name == value_node.value.id
                        ):
                            modules_with_all.append(module_name)
    return modules_with_all


def get_dunder_all_from_stub(package_path: str) -> List[str]:
    """
    Extracts the `__all__` symbols from a type stub (`.pyi`) file,
    given a path to the package containing it.

    Parameters
    ----------
    package_path : str
        The file path of package of the type stub file to parse.

    Returns
    -------
    list of str
        A list of symbols defined in `__all__` in the type stub.

    Raises
    ------
    Exception
        If `__all__` could not be found in the specified file.

    Notes
    -----
    This function specifically looks for assignments to `__all__` in the type
    stub file and returns a list of the symbols declared in it.
    """
    tree = ast.parse((Path(package_path) / "__init__.pyi").read_text())
    for node in tree.body:
        if isinstance(node, ast.Assign):
            for target in node.targets:
                if isinstance(target, ast.Name) and target.id == "__all__":
                    if isinstance(node.value, ast.List):
                        return [
                            element.s
                            for element in node.value.elts
                            if isinstance(element, ast.Constant)
                        ]

    raise Exception("__all__ could not be found")


def get_eligible_package_paths() -> Iterable[str]:
    """
    Traverses the `hstrat` package directory and returns
    the path for every directory that contains an `__init__.pyi`,
    as these are the valid packages to run the following checks on.
    """
    for path, _, files in os.walk("hstrat"):
        if "__init__.pyi" in files:
            yield path


def _symbol_not_in_stub(mod_path: str, sym: str):
    """
    Helper function for below to format a message in the case where
    a symbol defined in the `__all__` of a module is not in the type stub.
    """
    return f"Package '{module_name_from_path(mod_path)}' includes '{sym}' but does not declare it in the type stub."


def _symbol_not_in_module(mod_path: str, sym: str):
    """
    Helper function for below to format a message in the case where
    a symbol defined in the `__all__` of a type stub is not in the module.
    """
    return f"Type stub for '{module_name_from_path(mod_path)}' declares '__all__' that contains '{sym}' while the package itself does not contain it."


def check_accurate_all_declarations() -> Iterable[str]:
    """
    Check that the `__all__` from a package is the same as that
    defined in the type stub. Does this by reading the `__all__`
    from the type stub, and comparing it an `__all__` imported
    directly from the package.
    """
    for path in get_eligible_package_paths():
        type_stub_all = set(get_dunder_all_from_stub(path))
        module_all = set(getattr(import_from_path(path), "__all__"))
        for sym in type_stub_all - module_all:
            yield _symbol_not_in_module(path, sym)
        for sym in module_all - type_stub_all:
            yield _symbol_not_in_stub(path, sym)


def check_accurate_subpackage_star_imports() -> Iterable[str]:
    """
    If a package imports `__all__` from a subpackages, makes sure that
    that subpackage's real `__all__` is a subset of the `__all__` declared
    in the type stub of the importing package.
    """
    for path in get_eligible_package_paths():
        type_stub_all = get_dunder_all_from_stub(path)
        for subpackage in find_modules_with_all_references(path):
            subpackage_all = getattr(
                import_from_path(os.path.join(path, subpackage)), "__all__"
            )
            for sym in set(subpackage_all) - set(type_stub_all):
                yield f"Type stub for '{module_name_from_path(path)}' declares '__all__' that does not contain symbol '{sym}' from subpackage '{subpackage}'."


def check_accurate_flat_namespace() -> Iterable[str]:
    """
    Checks the flat namespace `hstrat.hstrat` and assures
    that the `__all__` defined in its `__init__.pyi` is accurate,
    containing all symbols from flattened submodules.
    """
    type_stub_all = get_dunder_all_from_stub(os.path.join("hstrat", "hstrat"))
    for subpackage in os.listdir("hstrat"):
        if (
            subpackage != "hstrat"
            and not subpackage.startswith("_")
            and (
                os.path.exists(
                    os.path.join("hstrat", subpackage, "__init__.pyi"),
                )
            )
        ):
            subpackage_all = getattr(
                import_from_path(os.path.join("hstrat", subpackage)), "__all__"
            )
            for sym in set(subpackage_all) - set(type_stub_all):
                yield f"Type stub for flat namespace 'hstrat.hstrat' does not contain symbol '{sym}' from '{subpackage}'."


if __name__ == "__main__":
    os.chdir(Path(__file__).parent.parent.parent.resolve())
    violations = [
        *check_accurate_all_declarations(),
        *check_accurate_subpackage_star_imports(),
        *check_accurate_flat_namespace(),
    ]
    print(f"{len(violations)} violations found.")
    for v in violations:
        print(v)

    exit_code = len(violations) > 0
    exit(exit_code)



from ..._impl import CalcMrcaUncertaintyRelExactFromAbs


# must inherit rather than assign due to failure of attribute lookup when
# pickling otherwise
class CalcMrcaUncertaintyRelExact(CalcMrcaUncertaintyRelExactFromAbs):
    pass



"""Retain as few strata as possible.

The nominal resolution policy only retains the most ancient (i.e., very
first) and most recent (i.e., last) strata. So, comparisons between two
columns under this policy will only be able to detect whether they share
any common ancestor and whether they are from the same organism (i.e., no
generations have elapsed since the MRCA). Thus, MRCA rank estimate
uncertainty scales as O(n) with respect to the greater number of strata deposited on either column.

Under the nominal resolution policy, the number of strata retained (i.e.,
space complexity) scales as O(1) with respect to the number of strata
deposited.
"""

from ...._auxiliary_lib import lazy_attach_stub

__getattr__, __dir__, __all__ = lazy_attach_stub(__name__, __file__)
del lazy_attach_stub



from typing import List

import pytest

from hstrat._auxiliary_lib import flag_last


@pytest.fixture
def sample_data() -> List[int]:
    return [1, 2, 3, 4, 5]


def test_empty_input():
    result = [*flag_last([])]
    assert result == []


def test_single_item_input():
    result = [*flag_last([1])]
    assert result == [(True, 1)]


def test_two_item_input():
    result = [*flag_last([1, 2])]
    assert result == [(False, 1), (True, 2)]


def test_even_input(sample_data):
    result = [*flag_last(sample_data[:4])]
    expected = [(False, 1), (False, 2), (False, 3), (True, 4)]
    assert result == expected


def test_odd_input(sample_data):
    result = [*flag_last(sample_data)]
    expected = [(False, 1), (False, 2), (False, 3), (False, 4), (True, 5)]
    assert result == expected


def test_iterator_input(sample_data):
    result = [*flag_last(iter(sample_data))]
    expected = [(False, 1), (False, 2), (False, 3), (False, 4), (True, 5)]
    assert result == expected


def test_mixed_type_input():
    result = [*flag_last([1, 2, "a"])]
    expected = [(False, 1), (False, 2), (True, "a")]
    assert result == expected



import typing


# adapted from https://stackoverflow.com/a/22490617
def cmp(a: typing.Any, b: typing.Any) -> int:
    """Compare two objects and return an integer indicating their relative
    order.

    Parameters
    ----------
    a : Any
        The first object to be compared.
    b : Any
        The second object to be compared.

    Returns
    -------
    int
        1 if `a` is greater than `b`, -1 if `a` is less than `b`, 0 otherwise.

    Example
    -------
    >>> cmp(1, 2)
    -1
    >>> cmp(2, 1)
    1
    >>> cmp(2, 2)
    0
    """
    return bool(a > b) - bool(a < b)



import logging
import math

from downstream import dataframe as dstream_dataframe
import polars as pl
import tqdm

from .._auxiliary_lib import (
    alifestd_make_empty,
    log_context_duration,
    log_memory_usage,
    render_polars_snapshot,
)
from ..phylogenetic_inference.tree._impl._build_tree_searchtable_cpp_impl_stub import (
    Records,
    collapse_dropped_unifurcations,
    extend_tree_searchtable_cpp_from_exploded,
    records_to_dict,
)


def _get_sole_bitwidth(df: pl.DataFrame) -> int:
    """Get dstream bitwidth value from DataFrame, ensuring it is unique."""
    assert not df.lazy().limit(1).collect().is_empty()
    if (
        not df.lazy()
        .filter(pl.col("dstream_value_bitwidth").diff() != pl.lit(0))
        .limit(1)
        .collect()
        .is_empty()
    ):
        raise NotImplementedError(
            "multiple differentia_bitwidths not yet supported",
        )
    return df.lazy().select("dstream_value_bitwidth").limit(1).collect().item()


def _build_records_chunked(
    df: pl.DataFrame, exploded_slice_size: int
) -> Records:
    """Build tree searchtable from DataFrame, exploding in chunks to reduce
    memory usage."""
    num_slices = math.ceil(len(df) / exploded_slice_size)
    logging.info(f"{len(df)=} {exploded_slice_size=} {num_slices=}")

    init_size = len(df) * 8
    records = Records(init_size)
    for i, df_slice in enumerate(df.iter_slices(exploded_slice_size)):
        logging.info(f"incorporating slice {i}/{num_slices}...")

        with log_context_duration(
            f"explode dstream records, slice {i}", logging.info
        ):
            long_df = dstream_dataframe.explode_lookup_unpacked(
                df_slice, value_type="uint64"
            )

        if i == 0:
            render_polars_snapshot(long_df, "exploded", logging.info)

        with log_context_duration(f"build tree, slice {i}", logging.info):
            extend_tree_searchtable_cpp_from_exploded(
                records,
                long_df["dstream_data_id"].to_numpy(),
                long_df["dstream_T"].to_numpy(),
                long_df["dstream_Tbar"].to_numpy(),
                long_df["dstream_value"].to_numpy(),
                tqdm.tqdm,
            )

        with log_context_duration(
            f"collapse dropped unifurcations, slice {i}", logging.info
        ):
            records = collapse_dropped_unifurcations(records)

        log_memory_usage(logging.info)

    return records


def surface_unpack_reconstruct(
    df: pl.DataFrame,
    exploded_slice_size: int = 1_000_000,
) -> pl.DataFrame:
    """Unpack dstream buffer and counter from genome data and construct an
    estimated phylogenetic tree for the genomes.

    Parameters
    ----------
    df : pl.DataFrame
        The input DataFrame containing packed data with required columns, one
        row per genome.

        Required schema:
            - 'data_hex' : pl.String
                - Raw genome data, with serialized dstream buffer and counter.
                - Represented as a hexadecimal string.
            - 'dstream_algo' : pl.Categorical
                - Name of downstream curation algorithm used.
                - e.g., 'dstream.steady_algo'
            - 'dstream_storage_bitoffset' : pl.UInt64
                - Position of dstream buffer field in 'data_hex'.
            - 'dstream_storage_bitwidth' : pl.UInt64
                - Size of dstream buffer field in 'data_hex'.
            - 'dstream_T_bitoffset' : pl.UInt64
                - Position of dstream counter field in 'data_hex'.
            - 'dstream_T_bitwidth' : pl.UInt64
                - Size of dstream counter field in 'data_hex'.
            - 'dstream_S' : pl.Uint32
                - Capacity of dstream buffer, in number of data items.

        Optional schema:
            - 'downstream_version' : pl.Categorical
                - Version of downstream library used to curate data items.
            - 'dstream_data_id' : pl.UInt64
                - Unique identifier for each data item.
                - If not provided, row number will be used as identifier.
            - 'downstream_exclude_exploded' : pl.Boolean
                - Should row be dropped after exploding unpacked data?
            - 'downstream_exclude_unpacked' : pl.Boolean
                - Should row be dropped after unpacking packed data?
            - 'downstream_validate_exploded' : pl.String, polars expression
                - Polars expression to validate exploded data.
            - 'downstream_validate_unpacked' : pl.String, polars expression
                - Polars expression to validate unpacked data.

    exploded_slice_size : int, default 1_000_000
        Number of rows to process at once. Lower values reduce memory usage.

    Returns
    -------
    pl.DataFrame
        The output DataFrame containing the estimated phylogenetic tree in
        alife standard format, with the following columns:

        - 'taxon_id' : pl.UInt64
            - Unique identifier for each taxon (RE alife standard format).
        - 'ancestor_id' : pl.UInt64
            - Unique identifier for ancestor taxon  (RE alife standard format).
        - 'origin_tme' : pl.UInt64
            - Num generations elapsed for ancestral differentia.
            - RE alife standard format.
            - a.k.a. "rank"
                - Corresponds to`dstream_Tbar` for inner nodes.
                - Corresponds `dstream_T` - 1 for leaf nodes
        - 'differentia_bitwidth' : pl.UInt64
            - Size of annotation differentiae, in bits.
            - Corresponds to `dstream_value_bitwidth`.
        - 'dstream_data_id' : pl.UInt64
            - Unique identifier for each genome in source dataframe
            - Set to source dataframe row index if not provided.

        User-defined columns, except some prefixed with 'downstream_' or
        'dstream_', will be forwarded from the input DataFrame.

        Note that the alife-standard `ancestor_list` column is not included in
        the output.
    """
    logging.info("beginning surface_unpack_reconstruct")
    log_memory_usage(logging.info)

    render_polars_snapshot(df, "packed", logging.info)

    # for simplicity, return early for this special case
    if df.lazy().limit(1).collect().is_empty():
        logging.info("empty input dataframe, returning empty result")
        return pl.from_pandas(alifestd_make_empty())

    logging.info("extracting differentia bitwidth...")
    df = df.with_columns(
        dstream_value_bitwidth=(
            pl.col("dstream_storage_bitwidth") // pl.col("dstream_S")
        ),
    )
    differentia_bitwidth = _get_sole_bitwidth(df)
    logging.info(f" - differentia bitwidth: {differentia_bitwidth}")

    with log_context_duration("unpack genome strings", logging.info):
        df = dstream_dataframe.unpack_data_packed(df)
    render_polars_snapshot(df, "unpacked", logging.info)

    logging.info("building tree searchtable chunkwise...")
    records = _build_records_chunked(df, exploded_slice_size)

    logging.info("converting records to dict...")
    records_dict = records_to_dict(records)

    logging.info("finalizing phylogeny dataframe...")

    phylo_df = pl.from_dict(
        records_dict,  # type: ignore
        schema={
            "dstream_data_id": pl.UInt64,
            "id": pl.UInt64,
            "ancestor_id": pl.UInt64,
            "differentia": pl.UInt64,
            "rank": pl.UInt64,
        },
    )
    del records
    phylo_df = phylo_df.with_columns(
        pl.lit(differentia_bitwidth)
        .alias("differentia_bitwidth")
        .cast(pl.UInt32),
    )

    logging.info("joining user-defined columns...")
    with log_context_duration("join user-defined columns", logging.info):
        df = df.select(
            pl.exclude("^dstream_.*$", "^downstream_.*$"),
            pl.col("dstream_data_id").cast(pl.UInt64),
        )
        joined_columns = set(df.columns) - set(phylo_df.columns)
        if joined_columns:
            logging.info(f" - {len(joined_columns)} column to join")
            logging.info(f" - joining columns: {[*joined_columns]}")
            phylo_df = phylo_df.join(df, on="dstream_data_id", how="left")
        else:
            logging.info(" - no columns to join, skipping")

    logging.info("surface_unpack_reconstruct complete")
    render_polars_snapshot(phylo_df, "reconstruction", logging.info)

    return phylo_df



import time

import more_itertools as mit
import pandas as pd

from hstrat._auxiliary_lib import log_context_duration


def test_log_context_duration():
    log = []
    with log_context_duration("my_test_context", logger=log.append):
        time.sleep(0.05)

    logged_message = mit.one(log)
    assert "my_test_context" in logged_message

    timing_records = {
        k: v
        for line in logged_message.splitlines()
        if line.startswith("!!!")
        for k, v in eval(line.strip().removeprefix("!!!")).items()
    }
    timing_results = pd.DataFrame(
        {
            "what": timing_records.keys(),
            "duration (s)": timing_records.values(),
        },
    ).astype(
        {
            "what": str,
            "duration (s)": float,
        },
    )
    assert len(timing_results) == 1
    assert timing_results.at[0, "what"] == "my_test_context"
    assert timing_results.at[0, "duration (s)"] > 0



import pandas as pd

from hstrat._auxiliary_lib import (
    alifestd_assign_contiguous_ids,
    alifestd_is_chronologically_ordered,
    alifestd_make_empty,
    alifestd_try_add_ancestor_id_col,
)


def test_is_chronologically_ordered_empty():
    df = alifestd_make_empty()
    df["origin_time"] = []
    assert alifestd_is_chronologically_ordered(df)


def test_is_chronologically_ordered_sexual1():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[333,44]",
                "[333,5]",
                "[333]",
                "[none]",
                "[333]",
                "[none]",
            ],
            "origin_time": [5, 5, 2, 1, 2, 0],
        }
    )

    for phylogeny_df_ in phylogeny_df, alifestd_assign_contiguous_ids(
        phylogeny_df
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            df_ = df.copy()
            assert alifestd_is_chronologically_ordered(df)
            assert alifestd_is_chronologically_ordered(
                df.astype({"origin_time": float})
            )
            assert df.equals(df_)


def test_is_chronologically_ordered_sexual2():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[333,44]",
                "[333,5]",
                "[333]",
                "[none]",
                "[333]",
                "[none]",
            ],
            "origin_time": [5, 5, 2, 1, 2, 2],
        }
    )

    for phylogeny_df_ in phylogeny_df, alifestd_assign_contiguous_ids(
        phylogeny_df
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            df_ = df.copy()
            assert alifestd_is_chronologically_ordered(
                df.astype({"origin_time": float})
            )
            assert df.equals(df_)


def test_is_chronologically_unordered_sexual1():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[333,44]",
                "[333,5]",
                "[333]",
                "[none]",
                "[333]",
                "[none]",
            ],
            "origin_time": [5, 5, 2, 1, 2, 6],
        }
    )

    for phylogeny_df_ in phylogeny_df, alifestd_assign_contiguous_ids(
        phylogeny_df
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            df_ = df.copy()
            assert not alifestd_is_chronologically_ordered(df)
            assert not alifestd_is_chronologically_ordered(
                df.astype({"origin_time": float})
            )
            assert df.equals(df_)


def test_is_chronologically_unordered_sexual2():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[333,44]",
                "[333,5]",
                "[333]",
                "[none]",
                "[333]",
                "[none]",
            ],
            "origin_time": [1, 5, 2, 1, 2, 2],
        }
    )

    for phylogeny_df_ in (
        phylogeny_df,
        alifestd_assign_contiguous_ids(phylogeny_df),
        alifestd_try_add_ancestor_id_col(phylogeny_df),
        alifestd_try_add_ancestor_id_col(
            alifestd_assign_contiguous_ids(phylogeny_df)
        ),
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            df_ = df.copy()
            assert not alifestd_is_chronologically_ordered(df)
            assert not alifestd_is_chronologically_ordered(
                df.astype({"origin_time": float})
            )
            assert df.equals(df_)


def test_is_chronologically_unordered_sexual3():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[333,44]",
                "[333,5]",
                "[333]",
                "[none]",
                "[333]",
                "[none]",
            ],
            "origin_time": [1, 5, 2, 3, 2, 2],
        }
    )

    for phylogeny_df_ in (
        phylogeny_df,
        alifestd_assign_contiguous_ids(phylogeny_df),
        alifestd_try_add_ancestor_id_col(phylogeny_df),
        alifestd_try_add_ancestor_id_col(
            alifestd_assign_contiguous_ids(phylogeny_df)
        ),
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            df_ = df.copy()
            assert not alifestd_is_chronologically_ordered(df)
            assert not alifestd_is_chronologically_ordered(
                df.astype({"origin_time": float})
            )
            assert df.equals(df_)


def test_is_chronologically_ordered_asexual1():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[44]",
                "[5]",
                "[333]",
                "[none]",
                "[1]",
                "[333]",
            ],
            "origin_time": [6, 5, 2, 1, 6, 2],
        }
    )

    for phylogeny_df_ in (
        phylogeny_df,
        alifestd_assign_contiguous_ids(phylogeny_df),
        alifestd_try_add_ancestor_id_col(phylogeny_df),
        alifestd_try_add_ancestor_id_col(
            alifestd_assign_contiguous_ids(phylogeny_df)
        ),
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            df_ = df.copy()
            assert alifestd_is_chronologically_ordered(df)
            assert alifestd_is_chronologically_ordered(
                df.astype({"origin_time": float})
            )
            assert df.equals(df_)


def test_is_chronologically_ordered_asexual2():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[44]",
                "[5]",
                "[333]",
                "[none]",
                "[1]",
                "[none]",
            ],
            "origin_time": [6, 5, 2, 1, 6, 2],
        }
    )

    for phylogeny_df_ in (
        phylogeny_df,
        alifestd_assign_contiguous_ids(phylogeny_df),
        alifestd_try_add_ancestor_id_col(phylogeny_df),
        alifestd_try_add_ancestor_id_col(
            alifestd_assign_contiguous_ids(phylogeny_df)
        ),
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            df_ = df.copy()
            assert alifestd_is_chronologically_ordered(
                df.astype({"origin_time": float})
            )
            assert df.equals(df_)


def test_is_chronologically_unordered_asexual1():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[333]",
                "[333]",
                "[333]",
                "[none]",
                "[333]",
                "[0]",
            ],
            "origin_time": [5, 5, 2, 1, 2, 4],
        }
    )

    for phylogeny_df_ in (
        phylogeny_df,
        alifestd_assign_contiguous_ids(phylogeny_df),
        alifestd_try_add_ancestor_id_col(phylogeny_df),
        alifestd_try_add_ancestor_id_col(
            alifestd_assign_contiguous_ids(phylogeny_df)
        ),
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            df_ = df.copy()
            assert not alifestd_is_chronologically_ordered(df)
            assert not alifestd_is_chronologically_ordered(
                df.astype({"origin_time": float})
            )
            assert df.equals(df_)


def test_is_chronologically_unordered_asexual2():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[44]",
                "[0]",
                "[333]",
                "[none]",
                "[333]",
                "[none]",
            ],
            "origin_time": [1, 5, 2, 1, 2, 2],
        }
    )

    for phylogeny_df_ in (
        phylogeny_df,
        alifestd_assign_contiguous_ids(phylogeny_df),
        alifestd_try_add_ancestor_id_col(phylogeny_df),
        alifestd_try_add_ancestor_id_col(
            alifestd_assign_contiguous_ids(phylogeny_df)
        ),
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            df_ = df.copy()
            assert not alifestd_is_chronologically_ordered(df)
            assert not alifestd_is_chronologically_ordered(
                df.astype({"origin_time": float})
            )
            assert df.equals(df_)


def test_is_chronologically_unordered_asexual3():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[333]",
                "[5]",
                "[44]",
                "[none]",
                "[333]",
                "[none]",
            ],
            "origin_time": [1, 5, 2, 3, 2, 2],
        }
    )

    for phylogeny_df_ in (
        phylogeny_df,
        alifestd_assign_contiguous_ids(phylogeny_df),
        alifestd_try_add_ancestor_id_col(phylogeny_df),
        alifestd_try_add_ancestor_id_col(
            alifestd_assign_contiguous_ids(phylogeny_df)
        ),
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            df_ = df.copy()
            assert not alifestd_is_chronologically_ordered(df)
            assert not alifestd_is_chronologically_ordered(
                df.astype({"origin_time": float})
            )
            assert df.equals(df_)



import numbers
import typing
import warnings

import alifedata_phyloinformatics_convert as apc
import pandas as pd

from ...._auxiliary_lib import (
    alifestd_is_chronologically_ordered,
    alifestd_reroot_at_id_asexual,
)
from ._estimate_origin_times import estimate_origin_times


def time_calibrate_tree(
    alifestd_df: pd.DataFrame,
    leaf_node_origin_times: typing.Dict[int, numbers.Number],
    negative_origin_time_correction_method: typing.Optional[str] = None,
) -> pd.DataFrame:
    """Time-calibrates a given tree by estimating the origin time of its nodes,
    working backward from phylogenetic depths of leaf nodes using branch_length
    attributes.

    Sets `origin_time` attributes on tree nodes, reroots the tree to the most
    ancient (chronologically) clade, and sets the `branch_length` of the new
    root node to reflect estimated generations elapsed between genesis and the
    MRCA.

    Parameters
    ----------
    tree : pd.DataFrame
        Phylogeny to calibrate, in alife standard format.
    leaf_node_origin_times : dict
        A dictionary that maps the `id` attributes of leaf nodes in the tree
        to their phylogenetic depths since genesis.
    negative_origin_time_correction_method
        : {"truncate", "shift", "rescale"}, optional
        How should negative origin time estimates be corrected?
    """

    if not len(alifestd_df):
        return alifestd_df

    assert "branch_length" in alifestd_df or "edge_length" in alifestd_df
    digraph = apc.alife_dataframe_to_networkx_digraph(
        alifestd_df,
        setup_edge_lengths=True,
    )
    graph = digraph.to_undirected()
    node_origin_times = estimate_origin_times(graph, leaf_node_origin_times)

    new_root_ids = [
        node
        for node in graph.nodes
        # older than or same age as neighbors
        if not any(
            node_origin_times[neighbor] < node_origin_times[node]
            for neighbor in graph.neighbors(node)
        )
    ]
    assert new_root_ids  # guarantee reroot for consistent output columns
    for new_root_id in sorted(
        new_root_ids, key=lambda x: (node_origin_times[x], x), reverse=True
    ):
        alifestd_df = alifestd_reroot_at_id_asexual(
            alifestd_df.drop(
                ["branch_length", "edge_length"],
                axis="columns",
                errors="ignore",
            ),
            new_root_id,
            mutate=True,
        )

    alifestd_df["origin_time"] = alifestd_df["id"].map(node_origin_times)

    # could also shift and rescale or just shift,
    # maybe make kwarg option in future
    # for now, just truncate
    min_origin_time = alifestd_df["origin_time"].min()
    if min_origin_time < 0:
        if negative_origin_time_correction_method is None:
            warnings.warn(
                f"""Negative origin time(s) estimated with minimum {
                    min_origin_time
                }; truncating negative origin time(s) to zero. """
                "Set negative_origin_time_correction_method to silence this "
                "warning."
            )
            negative_origin_time_correction_method = "truncate"

        if negative_origin_time_correction_method == "truncate":
            alifestd_df.loc[alifestd_df["origin_time"] < 0, "origin_time"] = 0
        elif negative_origin_time_correction_method == "rescale":
            scale_to = alifestd_df["origin_time"].max()
            shift_by = -min_origin_time
            alifestd_df["origin_time"] += shift_by
            alifestd_df["origin_time"] *= scale_to / (scale_to + shift_by)
        elif negative_origin_time_correction_method == "shift":
            shift_by = -alifestd_df["origin_time"].min()
            alifestd_df["origin_time"] += shift_by
        else:
            assert False, negative_origin_time_correction_method

    assert alifestd_is_chronologically_ordered(alifestd_df)

    return alifestd_df



from collections import deque
import typing

import anytree

from ._AnyTreeFastLevelOrderIter import AnyTreeFastLevelOrderIter
from ._anynode_deepcopy_except_neighbors import (
    anynode_deepcopy_except_neighbors,
)


def anytree_iterative_deepcopy(
    node: anytree.AnyNode,
    progress_wrap: typing.Callable = lambda x: x,
) -> anytree.AnyNode:
    """Create a deep copy of an anytree tree without recursing over the tree
    structure.

    Parameters
    ----------
    node : anytree.AnyNode
        The root node of the anytree tree or subtree to copy.

    Returns
    -------
    anytree.AnyNode
        A new tree with the same structure and attributes as the original tree,
        but with no shared objects.
    """
    parent_from_deque = deque([node.parent])
    parent_to_deque = deque([None])

    for node_from in progress_wrap(AnyTreeFastLevelOrderIter(node)):

        while node_from.parent is not parent_from_deque[-1]:
            parent_from_deque.pop()
            parent_to_deque.pop()

        assert parent_from_deque[-1] is node_from.parent

        node_to = anynode_deepcopy_except_neighbors(node_from)
        # assigning to parent casues expensive __check_loop safety check
        # use internal __attach method instead
        node_to._NodeMixin__attach(parent_to_deque[-1])

        parent_from_deque.appendleft(node_from)
        parent_to_deque.appendleft(node_to)

    return node_to.root  # anytree uses a non-recursive implemenation



import pandas as pd

from ._alifestd_mark_ancestor_origin_time_asexual import (
    alifestd_mark_ancestor_origin_time_asexual,
)


def alifestd_mark_origin_time_delta_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """Add columns `origin_time_delta` and `ancestor_origin_time`.

    Dataframe must provide column `origin_time`.

    A topological sort will be applied if `phylogeny_df` is not topologically
    sorted. Reindexing may be applied.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """
    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_mark_ancestor_origin_time_asexual(
        phylogeny_df, mutate=True
    )

    phylogeny_df["origin_time_delta"] = (
        phylogeny_df["origin_time"] - phylogeny_df["ancestor_origin_time"]
    )

    return phylogeny_df



import sys
import typing

from bitarray import util as bitarray_util
import numpy as np
import typing_extensions

from .._auxiliary_lib import iter_slices, zip_strict
from ..genome_instrumentation import HereditaryStratum


def _make_buffer_bitarray(
    strata: typing.Iterable[HereditaryStratum],
    differentia_bit_width: int,
    always_omit_num_padding_bits_header: bool,
) -> typing_extensions.Buffer:
    try:
        len(strata)
    except TypeError:
        strata = [*strata]

    needs_header = (
        bool(differentia_bit_width % 8)
        and not always_omit_num_padding_bits_header
    )
    buffer_size = 8 * needs_header + len(strata) * differentia_bit_width
    buffer = bitarray_util.zeros(buffer_size)

    # use first byte to tell how many padding bits if any are required
    if needs_header:
        num_jagged_bits = len(buffer) % 8
        num_padding_bits = (8 - num_jagged_bits) % 8
        header = bitarray_util.int2ba(
            num_padding_bits,
            length=8,
        )
        buffer[:8] |= header

    for slice_, stratum in zip_strict(
        iter_slices(len(buffer), differentia_bit_width, needs_header * 8),
        strata,
    ):
        buffer[slice_] |= bitarray_util.int2ba(
            stratum.GetDifferentia(), length=differentia_bit_width
        )

    return buffer.tobytes()


def _make_buffer_numpy(
    strata: typing.Iterable[HereditaryStratum],
    differentia_bit_width: int,
) -> typing_extensions.Buffer:
    buffer = np.fromiter(
        (s.GetDifferentia() for s in strata),
        dtype=eval(f"np.uint{differentia_bit_width}"),
    )
    if {
        ">": "big",
        "<": "little",
        "=": sys.byteorder,
        "|": "not applicable",
    }[buffer.dtype.byteorder] == "little":
        buffer.byteswap(inplace=True)
    return buffer.tobytes()


def pack_differentiae_bytes(
    strata: typing.Iterable[HereditaryStratum],
    differentia_bit_width: int,
    always_omit_num_padding_bits_header: bool = False,
) -> typing_extensions.Buffer:
    """Pack a sequence of differentiae together into a compact
    representation.

    Returns a byte buffer concatenation of diffferentiae. If
    `differentia_bit_width` is not an even byte multiple, the first encoded
    byte tells how many empty padding bits, if any, were placed at the end of
    the concatenation in order to align the bitstring end to byte boundaries.
    """
    strata = [*strata]
    if differentia_bit_width in (8, 16, 32, 64):
        return _make_buffer_numpy(strata, differentia_bit_width)
    else:
        return _make_buffer_bitarray(
            strata, differentia_bit_width, always_omit_num_padding_bits_header
        )



import random

import numpy as np


def seed_random(seed: int) -> bool:
    """Seed all random sources used by hstrat library.

    Ensures reproducible execution.
    """
    random.seed(seed)
    np.random.seed(seed)



import typing


# adapted from https://stackoverflow.com/a/22115957
def unzip(
    iterable: typing.Iterable[typing.Tuple],
) -> typing.Iterator[typing.Tuple]:
    """Unzip the input iterable into an iterator of tuples where the nth output
    tuple contains the nth elements of each tuple in the input iterable.

    Parameters
    ----------
    iterable : typing.Iterable[typing.Tuple]
        An iterable of tuples to unzip.

    Returns
    -------
    typing.Iterator[typing.Tuple]
        An iterator of tuples where each tuple contains the nth elements from
        each of the tuples in the input iterable.

    Examples
    --------
    >>> list(unzip([(1, 2), (3, 4), (5, 6)]))
    [(1, 3, 5), (2, 4, 6)]
    Here, the first output tuple contains the first elements from each tuple in
    the input iterable, i.e. (1, 3, 5), while the second output tuple contains
    the second elements, i.e. (2, 4, 6).

    >>> list(unzip([(1, 2, 3), (4, 5, 6)]))
    [(1, 4), (2, 5), (3, 6)]
    Here, the first output tuple contains the first elements from each tuple in
    the input iterable, i.e. (1, 4), while the second output tuple contains the
    second elements, i.e. (2, 5), and the third output tuple contains the third
    elements, i.e. (3, 6).

    >>> list(unzip([(1,), (2,), (3,)]))
    [(1, 2, 3)]
    Here, the first output tuple contains the first and only element from each
    tuple in the input iterable, i.e. (1, 2, 3).
    """
    return zip(*iterable)



import numpy as np

from hstrat._auxiliary_lib import min_array_dtype


def test_single_value():
    arr = np.array([5])
    expected_dtype = np.dtype("uint8")
    assert min_array_dtype(arr) == expected_dtype
    assert (arr.astype(expected_dtype) == arr).all()


def test_single_value_negative():
    arr = np.array([-5])
    expected_dtype = np.dtype("int8")
    assert min_array_dtype(arr) == expected_dtype
    assert (arr.astype(expected_dtype) == arr).all()


def test_positive_values():
    arr = np.array([5, 10, 100])
    expected_dtype = np.dtype("uint8")
    assert min_array_dtype(arr) == expected_dtype
    assert (arr.astype(expected_dtype) == arr).all()


def test_mixed_values1():
    arr = np.array([-5, 0, 10])
    expected_dtype = np.dtype("int8")
    assert min_array_dtype(arr) == expected_dtype
    assert (arr.astype(expected_dtype) == arr).all()


def test_mixed_values2():
    arr = np.array([5, 0, -255])
    expected_dtype = np.dtype("int16")
    assert min_array_dtype(arr) == expected_dtype
    assert (arr.astype(expected_dtype) == arr).all()


def test_large_values():
    arr = np.array([10000, 100000, 1000000])
    expected_dtype = np.dtype("uint32")
    assert min_array_dtype(arr) == expected_dtype
    assert (arr.astype(expected_dtype) == arr).all()


def test_large_mixed_values():
    arr = np.array([-10000, 100000, 1000000])
    expected_dtype = np.dtype("int32")
    assert min_array_dtype(arr) == expected_dtype
    assert (arr.astype(expected_dtype) == arr).all()


def test_float_values():
    arr = np.array([-0.5, 1.0, 1.5])
    expected_dtype = np.dtype("float16")
    assert min_array_dtype(arr) == expected_dtype
    assert (arr.astype(expected_dtype) == arr).all()


def test_large_float_values():
    arr = np.array([0.5, 1.0e100, 1.5])
    expected_dtype = np.dtype("float64")
    assert min_array_dtype(arr) == expected_dtype
    assert (arr.astype(expected_dtype) == arr).all()


def test_complex_values():
    arr = np.array([1 + 2j, 3 + 4j, 5 + 6j])
    expected_dtype = np.dtype("complex64")
    assert min_array_dtype(arr) == expected_dtype
    assert (arr.astype(expected_dtype) == arr).all()


def test_empty_array():
    arr = np.array([])
    expected_dtype = np.dtype("uint8")
    assert min_array_dtype(arr) == expected_dtype
    assert (arr.astype(expected_dtype) == arr).all()



import typing


def reversed_enumerate(
    sequence: typing.Sequence,
) -> typing.Iterable[typing.Tuple[int, object]]:
    """Enumerate a sequence in reverse order.

    This function yields tuples containing the index of each element in the
    sequence and the element itself, with the index values starting from
    the last index in the sequence and counting down to zero.

    Parameters
    ----------
    sequence : typing.Sequence
        The sequence to enumerate.

        Must be sliceable.

    Yields
    -------
    Tuple[int, object]
        A tuple containing the index and element for each item in the
        sequence, in reverse order.

    Examples
    --------
    >>> l = "star"
    >>> list(reversed_enumerate(l))
    [(3, "r"), (2, "a"), (1, "t"), (0, "s")]
    """
    # numba jit doesn't support yield from or reversed (fully)
    for t in zip(range(len(sequence) - 1, -1, -1), sequence[::-1]):
        yield t



import dendropy as dp

from ._load_dendropy_tree import load_dendropy_tree


def setup_dendropy_tree(path: str) -> dp.Tree:
    tree = load_dendropy_tree(path)

    for node in tree:
        node.edge.length = 1

    for idx, node in enumerate(tree.leaf_node_iter()):
        node.taxon = tree.taxon_namespace.new_taxon(label=str(idx))

    tree.update_bipartitions(
        suppress_unifurcations=False,
        collapse_unrooted_basal_bifurcation=False,
    )

    return tree



from hstrat._auxiliary_lib import capitalize_n


def test_capitalize_n_empty_string():
    assert capitalize_n("", 0) == ""
    assert capitalize_n("", 1) == ""
    assert capitalize_n("", 2) == ""


def test_capitalize_n_capitalize_all():
    assert capitalize_n("hello world", 11) == "HELLO WORLD"
    assert capitalize_n("goodbye", 7) == "GOODBYE"
    assert capitalize_n("testing", 7) == "TESTING"
    assert capitalize_n("12345", 5) == "12345"


def test_capitalize_n_capitalize_none():
    assert capitalize_n("hello world", 0) == "hello world"
    assert capitalize_n("goodbye", 0) == "goodbye"
    assert capitalize_n("testing", 0) == "testing"


def test_capitalize_n_capitalize_partial():
    assert capitalize_n("hello world", 5) == "HELLO world"
    assert capitalize_n("goodbye", 2) == "GOodbye"
    assert capitalize_n("testing", 3) == "TESting"


def test_capitalize_n_capitalize_negative():
    assert capitalize_n("hello world", -5) == "hello WORLD"
    assert capitalize_n("testing", -1) == "testinG"



import copy

import pytest

from hstrat import hstrat


@pytest.mark.parametrize(
    "differentia_bit_width", [1, 2, 7, 8, 16, 32, 64, 129]
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=2
        ),
    ],
)
def test_init_and_getters(differentia_bit_width, retention_policy):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=differentia_bit_width,
        stratum_retention_policy=retention_policy,
    )
    for i in range(100):

        specimen = hstrat.col_to_specimen(column)
        assert specimen.GetNumStrataDeposited() == i + 1
        assert (
            specimen.GetStratumDifferentiaBitWidth()
            == column.GetStratumDifferentiaBitWidth()
        )
        assert specimen.GetNumStrataRetained() == column.GetNumStrataRetained()
        assert [*specimen.GetDifferentiaVals()] == [
            *column.IterRetainedDifferentia()
        ]
        assert [*specimen.GetRankIndex()] == [*column.IterRetainedRanks()]

        assert [*specimen.GetData()] == [*column.IterRetainedDifferentia()]
        assert [*specimen.GetData().index] == [*column.IterRetainedRanks()]

        for index in range(column.GetNumStrataRetained()):
            assert specimen.GetRankAtColumnIndex(
                index
            ) == column.GetRankAtColumnIndex(index)

        assert [*specimen.IterRetainedRanks()] == [*column.IterRetainedRanks()]
        assert [*specimen.IterRetainedDifferentia()] == [
            *column.IterRetainedDifferentia()
        ]

        assert (
            specimen.GetNumDiscardedStrata() == column.GetNumDiscardedStrata()
        )
        assert specimen.HasDiscardedStrata() == column.HasDiscardedStrata()

        column.DepositStratum()


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=2
        ),
    ],
)
def test_IterRankDifferentiaZip(retention_policy):
    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
    )

    for __ in range(100):
        c1.DepositStratum()
        s1 = hstrat.col_to_specimen(c1)
        assert [*s1.IterRankDifferentiaZip()] == [
            *zip(s1.IterRetainedRanks(), s1.IterRetainedDifferentia())
        ]
        iter_ = s1.IterRankDifferentiaZip(copyable=True)
        iter_copy = copy.copy(iter_)
        next(iter_copy)
        assert [*iter_copy] == [
            *zip(s1.IterRetainedRanks(), s1.IterRetainedDifferentia())
        ][1:]
        assert [*iter_] == [
            *zip(s1.IterRetainedRanks(), s1.IterRetainedDifferentia())
        ]



import itertools as it

import pytest

from hstrat import hstrat


def test_CalcDefinitiveMaxRankOfLastRetainedCommonalityWith_specimen():
    column = hstrat.HereditaryStratigraphicColumn()
    column2 = hstrat.HereditaryStratigraphicColumn()
    column.DepositStrata(100)

    child1 = column.CloneDescendant()
    child2 = column.CloneDescendant()

    assert hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
        hstrat.col_to_specimen(column), hstrat.col_to_specimen(column2)
    ) == hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
        column, column2
    )

    assert hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
        hstrat.col_to_specimen(column), hstrat.col_to_specimen(column)
    ) == hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
        column, column
    )

    assert hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
        hstrat.col_to_specimen(column), hstrat.col_to_specimen(child1)
    ) == hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
        column, child1
    )

    assert hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
        hstrat.col_to_specimen(child1), hstrat.col_to_specimen(child2)
    ) == hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
        child1, child2
    )


@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 2, 8],
)
def test_CalcDefinitiveMaxRankOfLastRetainedCommonalityWith1(
    ordered_store,
    differentia_width,
):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=differentia_width,
        stratum_ordered_store=ordered_store,
    )

    column.DepositStrata(100)

    offspring1 = column.CloneDescendant()
    offspring2 = column.CloneDescendant()

    for c1, c2 in it.combinations([column, offspring1, offspring2], 2):
        if differentia_width == 64:
            assert (
                hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
                    c1,
                    c2,
                )
                == column.GetNumStrataDeposited() - 1
            )
            assert (
                hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
                    c2,
                    c1,
                )
                == column.GetNumStrataDeposited() - 1
            )
        else:
            assert (
                hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
                    c1,
                    c2,
                )
                >= column.GetNumStrataDeposited() - 1
            )
            assert (
                hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
                    c2,
                    c1,
                )
                >= column.GetNumStrataDeposited() - 1
            )

    for c in [column, offspring1, offspring2]:
        assert (
            hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
                c, c
            )
            == c.GetNumStrataDeposited() - 1
        )

    offspring1.DepositStrata(100)
    offspring2.DepositStrata(100)

    for c1, c2 in it.combinations([column, offspring1, offspring2], 2):
        if differentia_width == 64:
            assert (
                hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
                    c1,
                    c2,
                )
                == column.GetNumStrataDeposited() - 1
            )
            assert (
                hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
                    c2,
                    c1,
                )
                == column.GetNumStrataDeposited() - 1
            )
        else:
            assert (
                hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
                    c1,
                    c2,
                )
                >= column.GetNumStrataDeposited() - 1
            )
            assert (
                hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
                    c2,
                    c1,
                )
                >= column.GetNumStrataDeposited() - 1
            )

    for c in [column, offspring1, offspring2]:
        assert (
            hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
                c, c
            )
            == c.GetNumStrataDeposited() - 1
        )


@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_CalcDefinitiveMaxRankOfLastRetainedCommonalityWith2(ordered_store):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=1,
    )

    while True:
        offspring1 = column.CloneDescendant()
        offspring2 = column.CloneDescendant()
        res = hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
            offspring1,
            offspring2,
        )
        if res is not None and res > column.GetNumStrataDeposited() - 1:
            assert (
                hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
                    offspring2,
                    offspring1,
                )
                == res
            )
            break


@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_CalcDefinitiveMaxRankOfLastRetainedCommonalityWith3(ordered_store):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=64,
        stratum_retention_policy=hstrat.nominal_resolution_algo.Policy(),
    )

    column.DepositStrata(100)

    offspring1 = column.CloneDescendant()
    offspring2 = column.CloneDescendant()

    offspring1.DepositStrata(100)
    offspring2.DepositStrata(100)

    for c1, c2 in it.combinations([column, offspring1, offspring2], 2):
        assert c1.GetNumStrataRetained() == 2
        assert c2.GetNumStrataRetained() == 2

        assert (
            hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
                c1,
                c2,
            )
            == 0
        )
        assert (
            hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
                c2,
                c1,
            )
            == 0
        )


@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_CalcDefinitiveMaxRankOfLastRetainedCommonalityWith4(ordered_store):
    def do_once():
        c1 = hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=1,
        )
        c2 = hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=1,
        )

        res = hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
            c1,
            c2,
        )
        assert (
            hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
                c2,
                c1,
            )
            == res
        )
        assert res in (None, 0)

        if res is None:
            c1.DepositStrata(100)
            c2.DepositStrata(100)
            res = hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
                c1, c2
            )
            assert res is None
            assert (
                hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
                    c2,
                    c1,
                )
                is None
            )
        elif res == 0:
            while True:
                c1_ = c1.CloneDescendant()
                c2_ = c2.CloneDescendant()
                if (
                    hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
                        c1_, c2_
                    )
                    == 1
                ):
                    break

        return res

    while do_once() is not None:
        pass

    while do_once() != 0:
        pass


@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_CalcDefinitiveMaxRankOfLastRetainedCommonalityWith5(ordered_store):
    def do_once():
        c1 = hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=1,
            stratum_retention_policy=hstrat.fixed_resolution_algo.Policy(2),
        )
        c2 = hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=1,
            stratum_retention_policy=hstrat.fixed_resolution_algo.Policy(2),
        )

        res = hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
            c1,
            c2,
        )
        assert (
            hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
                c2,
                c1,
            )
            == res
        )
        assert res in (None, 0)

        if res is None:
            c1.DepositStrata(100)
            c2.DepositStrata(100)
            res = hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
                c1, c2
            )
            assert res is None
            assert (
                hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
                    c2,
                    c1,
                )
                is None
            )
        elif res == 0:
            while True:
                c1_ = c1.CloneDescendant()
                c2_ = c2.CloneDescendant()
                c1_.DepositStratum()
                c2_.DepositStratum()
                if (
                    hstrat.calc_definitive_max_rank_of_last_retained_commonality_between(
                        c1_, c2_
                    )
                    == 2
                ):
                    break

        return res

    while do_once() is not None:
        pass

    while do_once() != 0:
        pass



"""Visualizations for stratum retention policies."""

from . import animate, ascii, plot
from .._auxiliary_lib import lazy_attach

__getattr__, __dir__, __all__ = lazy_attach(
    __name__,
    submodules=["animate", "ascii", "plot"],
    submod_attrs={
        "animate": animate.__all__,
        "ascii": ascii.__all__,
        "plot": plot.__all__,
    },
    should_launder=[].__contains__,
)
del lazy_attach



import itertools as it
import os

import alifedata_phyloinformatics_convert as apc
import dendropy as dp
import pandas as pd
import pytest

from hstrat import hstrat
from hstrat._auxiliary_lib import pairwise, zip_strict

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "tree",
    [
        dp.Tree(),
        dp.Tree.get(
            path=f"{assets_path}/grandchild_and_aunt.newick", schema="newick"
        ),
        dp.Tree.get(
            path=f"{assets_path}/grandchild_and_auntuncle.newick",
            schema="newick",
        ),
        dp.Tree.get(path=f"{assets_path}/grandchild.newick", schema="newick"),
        dp.Tree.get(
            path=f"{assets_path}/grandtriplets_and_aunt.newick",
            schema="newick",
        ),
        dp.Tree.get(
            path=f"{assets_path}/grandtriplets_and_auntuncle.newick",
            schema="newick",
        ),
        dp.Tree.get(
            path=f"{assets_path}/grandtriplets.newick", schema="newick"
        ),
        dp.Tree.get(
            path=f"{assets_path}/grandtwins_and_aunt.newick", schema="newick"
        ),
        dp.Tree.get(
            path=f"{assets_path}/grandtwins_and_auntuncle.newick",
            schema="newick",
        ),
        dp.Tree.get(path=f"{assets_path}/grandtwins.newick", schema="newick"),
        dp.Tree.get(path=f"{assets_path}/justroot.newick", schema="newick"),
        dp.Tree.get(path=f"{assets_path}/triplets.newick", schema="newick"),
        dp.Tree.get(path=f"{assets_path}/twins.newick", schema="newick"),
        apc.alife_dataframe_to_dendropy_tree(
            pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        ),
        apc.alife_dataframe_to_dendropy_tree(
            pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        ),
        apc.alife_dataframe_to_dendropy_tree(
            pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
        ),
    ],
)
def test_perfect_tracking(tree):

    # setup tree
    for loc, node in enumerate(tree):
        node.taxon = tree.taxon_namespace.new_taxon(label=str(loc))

    tree.update_bipartitions(
        suppress_unifurcations=False,
        collapse_unrooted_basal_bifurcation=False,
    )

    handle_population = [
        hstrat.PerfectBacktrackHandle(
            data={
                "taxon_label": tree.seed_node.taxon.label,
            },
        )
    ]
    for (
        (ancestor_level, ancestor_level_nodes),
        (descendant_level, descendant_level_nodes),
    ) in pairwise(
        (
            (level, [*group])
            for level, group in it.groupby(
                tree.levelorder_node_iter(),
                lambda node: node.level(),
            )
        )
    ):
        ancestor_handle_lookup = {
            id(ancestor_node): ancestor_handle
            for (ancestor_node, ancestor_handle) in zip_strict(
                ancestor_level_nodes, handle_population
            )
        }

        handle_population = [
            ancestor_handle_lookup[
                id(descendant_node.parent_node)
            ].CreateDescendant(
                data={"taxon_label": descendant_node.taxon.label},
            )
            for descendant_node in descendant_level_nodes
        ]

    alifedata_df = hstrat.compile_perfect_backtrack_phylogeny(
        handle_population,
    )
    assert len(alifedata_df) <= sum(1 for __ in tree)
    assert len(alifedata_df) == len(alifedata_df["id"].unique())

    tracked_tree = apc.alife_dataframe_to_dendropy_tree(
        alifedata_df,
    )
    tracked_tree.is_rooted = tree.is_rooted

    max_level = max((node.level() for node in tree.leaf_node_iter()))
    extant_subtree = tree.extract_tree(
        node_filter_fn=lambda node: node.level() == max_level
    )
    assert len(tracked_tree) == len(extant_subtree)
    assert len(extant_subtree)

    tracked_tree.migrate_taxon_namespace(extant_subtree.taxon_namespace)
    assert tracked_tree.taxon_namespace is extant_subtree.taxon_namespace
    for node in tracked_tree:
        assert node.taxon in extant_subtree.taxon_namespace

    assert (
        dp.calculate.treecompare.symmetric_difference(
            extant_subtree,
            tracked_tree,
            is_bipartitions_updated=True,
        )
        == 0
    )
